# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module to maintain a run of a network node.
"""

import weakref
from collections import OrderedDict

from .basenoderun import BaseNodeRun
from .inputoutputrun import BaseInputRun, InputRun, OutputRun
from .job import Job, JobState
from .. import exceptions, resources
from ..core.samples import SampleId, SampleIndex, SampleItem, SampleState
from ..core.resourcelimit import ResourceLimit
from ..core.tool import Tool
from ..datatypes import DataType
from ..helpers import log
from ..planning.inputgroupcombiner import DefaultInputGroupCombiner, MergingInputGroupCombiner
from ..planning.node import InputGroup

__all__ = ['NodeRun']


class NodeRun(BaseNodeRun):
    """
    The class encapsulating a node in the network. The node is responsible
    for setting and checking inputs and outputs based on the description
    provided by a tool instance.
    """
    __dataschemafile__ = 'NodeRun.schema.json'

    _InputType = InputRun
    _OutputType = OutputRun
    _JobType = Job

    def __init__(self, node, parent):
        """
        Instantiate a node.

        :param node: The node to base the noderun on
        :type node: :py:class:`Tool <fastr.planning.node.NodeRun>`
        :param parent: the parent network of the node
        :type parent: :py:class:`Network <fastr.planning.network.Network>`
        :return: the newly created NodeRun
        """
        super(NodeRun, self).__init__()

        self._tool = node.tool
        self._parent = weakref.ref(parent)
        self._id = node.id
        self.nodegroup = node.nodegroup
        self.drained = False

        inputlist = []
        outputlist = []

        # Create Input/Output objects
        for input_ in node.inputs.values():
            # It can happen that this has been done by a subclass, check first
            inputobj = self._InputType(self, input_)
            inputlist.append((inputobj.id, inputobj))

        for output in node.outputs.values():
            # It can happen that this has been done by a subclass, check first
            outputobj = self._OutputType(self, output)
            outputlist.append((outputobj.id, outputobj))

        self.inputs = OrderedDict(inputlist)
        self.outputs = OrderedDict(outputlist)

        self._input_groups = OrderedDict()

        # Set the job requirements
        self._resources = node.resources

        # Set the flow control
        self.input_group_combiner = None
        self.merge_dimensions = node.merge_dimensions

        # Set the input groups
        self.update_input_groups()
        self.input_group_combiner.update()

        self.jobs = {}

        # Update Inputs and self (which calls Outputs)
        self.update()

    def __repr__(self):
        """
        Get a string representation for the NodeRun

        :return: the string representation
        :rtype: str
        """
        return str(self)

    def __str__(self):
        """
        Get a string version for the NodeRun

        :return: the string version
        :rtype: str
        """
        return "<{}: {}>".format(type(self).__name__, self.id)

    def __eq__(self, other):
        """Compare two Node instances with each other. This function ignores
        the parent and update status, but tests rest of the dict for equality.
        equality

        :param other: the other instances to compare to
        :type other: NodeRun
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, NodeRun):
            return NotImplemented

        dict_self = dict(vars(self))
        del dict_self['_parent']
        del dict_self['_status']
        del dict_self['_input_groups']
        del dict_self['jobs']
        del dict_self['input_group_combiner']

        dict_other = dict(vars(other))
        del dict_other['_parent']
        del dict_other['_status']
        del dict_other['_input_groups']
        del dict_other['jobs']
        del dict_other['input_group_combiner']

        return dict_self == dict_other

    def __getstate__(self):
        """
        Retrieve the state of the NodeRun

        :return: the state of the object
        :rtype dict:
        """
        state = super(NodeRun, self).__getstate__()

        # Make id prettier in the result
        state['id'] = self.id

        # Add the class of the NodeRun in question
        state['class'] = type(self).__name__

        # Get all input and output
        state['inputs'] = [x.__getstate__() for x in self.inputs.values()]
        state['outputs'] = [x.__getstate__() for x in self.outputs.values()]

        if self._tool is not None:
            state['tool'] = [self._tool.ns_id, str(self._tool.command['version'])]
        else:
            state['tool'] = None

        # Add resource requirements
        state['_resources'] = self._resources.__getstate__()
        state['merge_dimensions'] = self._merge_dimensions

        return state

    def __setstate__(self, state):
        """
        Set the state of the NodeRun by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        # Make sure the NodeRun classes are aligned (and warn if not so)
        if 'class' in state and state['class'] != type(self).__name__:
            log.warning('Attempting to set the state of a {} to {}'.format(
                state['class'],
                type(self).__name__
            ))

        self.jobs = None

        if not hasattr(self, '_input_groups'):
            self._input_groups = OrderedDict()

        if 'id' in state:
            self._id = state['id']

        if 'parent' in state:
            parent = state['parent']
            del state['parent']
        else:
            parent = None

        if state['tool'] is not None:
            self._tool = resources.tools[tuple(state['tool'])]
        else:
            self._tool = None

        # Create Input/Output objects
        inputlist = []
        for input_ in state['inputs']:
            if 'node' in input_:
                # Check if the expected NodeRun id matches our current id
                if input_['node'] != state['id']:
                    raise exceptions.FastrParentMismatchError('This Input has different parent node!')
                del input_['node']

            # It can happen that this has been done by a subclass, check first
            if not isinstance(input_, BaseInputRun):
                description = self.tool.inputs[input_['id']]
                inputobj = self._InputType(self, description)
                inputobj._node = self
                inputobj.__setstate__(input_)
            else:
                inputobj = input_
            inputlist.append((inputobj.id, inputobj))

        outputlist = []
        for output in state['outputs']:
            if '_node' in output:
                # Check if the expected NodeRun id matches our current id
                if output['_node'] != state['_id']:
                    raise exceptions.FastrParentMismatchError('This Input has different parent node!')
                del output['_node']

            # It can happen that this has been done by a subclass, check first
            if not isinstance(output, OutputRun):
                description = self.tool.outputs[output['id']]
                outputobj = self._OutputType(self, description)
                outputobj.__setstate__(output)
            else:
                outputobj = output
            outputlist.append((outputobj.id, outputobj))

        self.inputs = OrderedDict(inputlist)
        self.outputs = OrderedDict(outputlist)

        super(NodeRun, self).__setstate__(state)

        self._parent = None
        if parent is not None:
            self.parent = parent
        else:
            message = 'parent argument is None, need a parent Network to function!'
            raise exceptions.FastrValueError(message)

        self._resources = ResourceLimit()
        self._resources.__setstate__(state['resources'])
        self.merge_dimensions = state['merge_dimensions']

        self.update()

    @property
    def merge_dimensions(self):
        return self._merge_dimensions

    @merge_dimensions.setter
    def merge_dimensions(self, value):
        if isinstance(value, str):
            options = ['all', 'none']
            if value not in options:
                raise exceptions.FastrValueError('Invalid option {} given (valid options: {})'.format(value, options))
            self._merge_dimensions = value
            if value == 'none':
                self.input_group_combiner = DefaultInputGroupCombiner(self)
            elif value == 'all':
                self.input_group_combiner = MergingInputGroupCombiner(self, value)
        else:
            self._merge_dimensions = value
            self.input_group_combiner = MergingInputGroupCombiner(self, tuple(value))

    def update_input_groups(self):
        """
        Update all input groups in this node
        """
        input_groups = OrderedDict()

        for input_ in self.inputs.values():
            if input_.input_group not in input_groups:
                input_groups[input_.input_group] = InputGroup(self, input_.input_group)
            input_groups[input_.input_group][input_.id] = input_

        self._input_groups = input_groups

    @classmethod
    def createobj(cls, state, network=None):
        if 'parent' not in state:
            if network is not None:
                log.debug('Setting network to: {}'.format(network))
                state['parent'] = network
            else:
                log.debug('No network given for de-serialization')
        else:
            log.debug('Parent is already defined as: {}'.format(network))

        state = dict(state)

        return super(NodeRun, cls).createobj(state, network)

    @property
    def blocking(self):
        """
        Indicate that the results of this NodeRun cannot be determined without first executing the NodeRun, causing a
        blockage in the creation of jobs. A blocking Nodes causes the Chunk borders.
        """
        for output in self.outputs.values():
            if output.blocking:
                log.debug('Blocking because Output {} has cardinality {}'.format(output.fullid,
                                                                                       output.cardinality()))
                return True
        return False

    @property
    def dimnames(self):
        """
        Names of the dimensions in the NodeRun output. These will be reflected
        in the SampleIdList of this NodeRun.
        """
        if hasattr(self, '_dimnames') and self._dimnames is not None:
            return self._dimnames
        else:
            return self.input_group_combiner.dimnames

    @dimnames.setter
    def dimnames(self, value):
        if isinstance(value, str):
            value = value,

        if not isinstance(value, tuple) or not all(isinstance(x, str) for x in value):
            raise exceptions.FastrTypeError('Dimnames has to be a tuple of str!')

        log.warning('You are overriding the dimnames of a NodeRun, beware this is possible but not encouraged and can lead to strange results!')
        self._dimnames = value

    @dimnames.deleter
    def dimnames(self):
        del self._dimnames

    @property
    def fullid(self):
        """
        The full defining ID for the NodeRun inside the network
        """
        return '{}/nodelist/{}'.format(self.parent.fullid, self.id)

    @property
    def global_id(self):
        """
        The global defining ID for the Node from the main network (goes out
        of macro nodes to root network)
        """
        return '{}/nodelist/{}'.format(self.parent.global_id, self.id)

    @property
    def input_groups(self):
        """
        A list of input groups for this NodeRun. An input group is InputGroup
         object filled according to the NodeRun

        """
        return self._input_groups

    @property
    def outputsize(self):
        """
        Size of the outputs in this NodeRun
        """
        # Get sizes of all input groups
        return self.input_group_combiner.size

    @property
    def id(self):
        """
        The id of the NodeRun
        """
        return self._id

    @property
    def listeners(self):
        """
        All the listeners requesting output of this node, this means the
        listeners of all Outputs and SubOutputs
        """
        listeners = []
        for output in self.outputs.values():
            for listener in output.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        return listeners

    @property
    def name(self):
        """
        Name of the Tool the NodeRun was based on. In case a Toolless NodeRun was
        used the class name is given.
        """
        if hasattr(self, '_tool') and isinstance(self._tool, Tool):
            return self._tool.id
        else:
            return self.__class__.__name__

    @property
    def parent(self):
        """
        The parent network of this node.
        """
        return self._parent()

    @parent.setter
    def parent(self, value):
        """
        The parent network of this node. (setter)
        """
        if self._parent() is value:
            return  # Setting to same value doesn't do anything

        if self._parent() is not None:
            raise exceptions.FastrAttributeError('Cannot reset attribute once set')

        self._parent = weakref.ref(value)
        self._parent().add_node(self)

    @property
    def resources(self):
        """
        Number of cores required for the execution of this NodeRun
        """
        return self._resources

    @property
    def status(self):
        return self._status

    @property
    def tool(self):
        return self._tool

    def _execute(self):
        """
        Execute the node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        self.update(False, False)

        # Make sure a NodeRun is valid
        if not self.valid:
            message = 'NodeRun {} is not valid'.format(self.fullid)
            log.error(message)
            log.error('Messages:\n{}'.format('\n'.join(self.messages)))
            raise exceptions.FastrNodeNotValidError(message)
        input_groups = self.input_groups

        log.info('Generating jobs for node "{}" with dimensions: {}'.format(
            self.id,
            ', '.join('[{}: {}]'.format(name, size) for name, size in zip(self.dimnames, self.outputsize))
        ))

        for sample_index, sample_id, job_data, job_depends, failed_annotations, status in self.input_group_combiner:
            job = self.create_job(sample_id, sample_index, job_data, job_depends, status)

            if status == SampleState.MISSING:
                job.status = JobState.execution_skipped
                job.write()

            yield job

        log.debug('===== END execute_node =====')
        self.drained = True

    def execute(self):
        """
        Execute the node and create the jobs that need to run

        :return: list of jobs to run
        :rtype: list of :py:class:`Jobs <fastr.execution.job.Job>`
        """
        yield self._execute()

    def set_result(self, job, failed_annotation):
        """
        Incorporate result of a job into the NodeRun.

        :param Type job: job of which the result to store
        :param failed_annotation: A set of annotations, None if no errors else containing a tuple describing the errors
        """
        sample_id, sample_index = job.sample_id, job.sample_index
        # Replace following code by node.set_data(job.output_data) ? or something like it?
        for output in self.outputs.values():
            if output.id not in job.output_data and len(output.listeners) > 0 and len(failed_annotation) == 0:
                error_message = 'Could not find required data for {} in {}!'.format(output.fullid, job.output_data)
                log.error(error_message)

        for output in self.outputs.values():
            # No Errors and No samples in output
            missing_input_data = any(x.status == SampleState.MISSING for x in job.input_arguments.values())

            if not failed_annotation and self.blocking:
                log.debug('>>>> >>>> No Errors and No samples in output in sample[{};{}]'.format(sample_id, sample_index))

                if output.id not in job.output_data:
                    # There is not data, skip this output, if this was a problem,
                    # a failure should have been detected anyways, but probably it
                    # was a non-required output
                    continue

                output_data = job.output_data[output.id]

                log.debug('Setting data for blocking node: {} sample: {} with annotation: {}'.format(output.fullid,
                                                                                                     sample_id,
                                                                                                     failed_annotation))

                output_values = tuple(job.get_deferred(output.id, c) for c in range(len(output_data)))

                log.debug('Setting collected for {} sample_id {} data: {}'.format(output.fullid,
                                                                                  sample_id,
                                                                                  output_values))

                status = SampleState.MISSING if missing_input_data else SampleState.VALID
                output[sample_id, sample_index] = SampleItem(sample_index,
                                                             sample_id,
                                                             OrderedDict({0: tuple(output_values)}),
                                                             {job},
                                                             failed_annotation,
                                                             status)
            # Errors and no samples
            elif failed_annotation and self.blocking:
                output_values = (job.get_deferred(output.id, 0),)

                log.debug('Setting data for blocking node: {} sample: {} with annotation: {}'.format(output.fullid,
                                                                                                     sample_id,
                                                                                                     failed_annotation))

                log.debug('>>>> >>>> Errors and No samples in output in sample[{};{}]'.format(sample_id, sample_index))
                status = SampleState.MISSING if missing_input_data else SampleState.FAILED
                output[sample_id, sample_index] = SampleItem(sample_index,
                                                             sample_id,
                                                             OrderedDict({0: tuple(output_values)}),
                                                             {job},
                                                             failed_annotation,
                                                             status)

                log.debug('$$ new annotation: {}'.format(output[sample_index].failed_annotations))
                # Errors and samples
            elif failed_annotation and not self.blocking:
                log.debug('>>>> >>>> Errors and samples in output in sample[{};{}]'.format(sample_id, sample_index))
                sample_item = output[sample_index]

                # Update failed annotations and status
                new_failed_annotation = set(sample_item.failed_annotations)
                new_failed_annotation.update(failed_annotation)
                new_status = SampleState.MISSING if sample_item.status == SampleState.MISSING else SampleState.FAILED

                new_item = sample_item.replace(
                    failed_annotations=new_failed_annotation,
                    status=new_status,
                )
                output[sample_index] = new_item
            else:
                log.debug(">>>> >>>> No errors and samples in output in sample[{};{}]".format(sample_id, sample_index))

            self.jobs[sample_id] = job

    def create_job(self, sample_id: SampleId, sample_index: SampleIndex, job_data,
                   job_dependencies, status, **kwargs):
        """
        Create a job based on the sample id, job data and job dependencies.

        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :param sample_index: the index of the corresponding sample
        :type sample_index: :py:class:`SampleIndex <fastr.core.sampleidlist.SampleIndex>`
        :param dict job_data: dictionary containing all input data for the job
        :param job_dependencies: other jobs that need to finish before this job can run
        :return: the created job
        :rtype: :py:class:`Job <fastr.execution.job.Job>`
        """
        log.info('Creating job for node {} sample id {!r}, index {!r}'.format(self.global_id, sample_id, sample_index))
        log.debug('Creating job for sample {} with data {}'.format(sample_id, job_data))

        # Get the arguments
        input_arguments, output_arguments = self._wrap_arguments(job_data, sample_id, sample_index)

        preferred_types = {output.id: output.preferred_types for output in self.outputs.values()}

        job = self._JobType(node=self,
                            sample_id=sample_id,
                            sample_index=sample_index,
                            input_arguments=input_arguments,
                            output_arguments=output_arguments,
                            hold_jobs=job_dependencies,
                            preferred_types=preferred_types,
                            **kwargs)
        self.jobs[sample_id] = job

        # Check which outputs are required or connected and set them
        if not self.blocking:
            for output in self.outputs.values():
                # Not that this always has to happen, as we need the samples
                # to possibly annotate errors later, even if the output will
                # not be used later because of a lack of listeners
                log.debug('Preparing output {}'.format(output.id))
                log.debug('Cardinality request: spec: {}, job_data: {}, and index: {}'.format(output.cardinality_spec,
                                                                                              job_data,
                                                                                              sample_index))
                cardinality = output.cardinality(sample_index, job_data)
                log.debug('Cardinality for {} is {}'.format(output.id, cardinality))
                if not isinstance(cardinality, int):
                    message = 'For execution cardinality should be an int, for output ' \
                              '{} we found {} (type {})'.format(output.id,
                                                                cardinality,
                                                                type(cardinality).__name__)
                    log.critical(message)
                    raise exceptions.FastrTypeError(message)

                value = tuple(job.get_deferred(output.id, cardinality_nr) for cardinality_nr in range(cardinality))

                output[sample_id] = SampleItem(sample_index,
                                               sample_id,
                                               {0: value},
                                               {job},
                                               set(),
                                               status)
        else:
            log.debug('Cannot determine blocking node output a priori! Needs to be collected afterwards!')

        return job

    def _wrap_arguments(self, job_data, sample_id, sample_index):
        """
        Wrap arguments into a list of tuples that the execution script can parse

        :param dict job_data: dictionary containing all input data for the job
        :param sample_id: the id of the corresponding sample
        :type sample_id: :py:class:`SampleId <fastr.core.sampleidlist.SampleId>`
        :return: the wrapped arguments in a tuple with the form ``(inputs, outputs)``
        """
        arguments = ({}, {})  # format is (input_args, output_args)
        for key, input_ in self.inputs.items():
            # Skip inputs that have no data
            if job_data[key] is None:
                if input_.default is not None:
                    # Insert the default data if present
                    job_data[key] = SampleItem(0, 'default', (input_.datatype(input_.default),))
                elif input_.required:
                    log.debug('Job data: {}'.format(job_data))
                    raise exceptions.FastrValueError('NodeRun "{}" is missing data for required Input "{}"'.format(self.id, input_.id))
                else:
                    continue

            arguments[0][key] = job_data[key]

        for key, output in self.outputs.items():
            if not output.automatic:
                cardinality = output.cardinality(key=sample_index, job_data=job_data)
            else:
                cardinality = None

            if output.required or len(output.listeners) > 0:
                requested = True
            else:
                requested = False

            log.debug('Cardinality to be used: {}'.format(cardinality))

            arguments[1][key] = {
                'id': key,
                'cardinality': cardinality if cardinality is not None else str(output.cardinality_spec),
                'datatype': output.resulting_datatype.id,
                'requested': requested
            }

        return arguments

    def get_sourced_nodes(self):
        """
        A list of all Nodes connected as sources to this NodeRun

        :return: list of all nodes that are connected to an input of this node
        """
        sourced_nodes = []
        for input_ in self.inputs.values():
            for sourced_node in input_.get_sourced_nodes():
                if sourced_node not in sourced_nodes:
                    sourced_nodes.append(sourced_node)
        return sourced_nodes

    def find_source_index(self, target_index, target, source):
        # If there are multiple input groups, select only part of index from
        # the inputgroup which source belongs to
        if len(self.input_groups) > 1:
            input_groups = self.input_groups
            mask = [True if ig.id == source.input_group else False for ig in input_groups.values() for _ in ig.size]
            target_index = tuple(k for k, m in zip(target_index, mask) if m)

        # Delegate to InputGroup to check mixing within InputGroup
        return self.input_groups[source.input_group].find_source_index(target_size=target.size,
                                                                       target_dimnames=target.dimnames,
                                                                       source_size=source.size,
                                                                       source_dimnames=source.dimnames,
                                                                       target_index=target_index)

    def _update(self, key, forward=True, backward=False):
        """
        Update the NodeRun information and validity of the NodeRun and propagate
         the update downstream. Updates inputs, input groups, outputsize and outputs.

        A NodeRun is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.planning.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized
        """
        # Make sure the Inputs and input groups are up to date
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        if backward:
            for sourced_node in self.get_sourced_nodes():
                sourced_node.update(key, False, backward)

        for input_ in list(self.inputs.values()):
            input_.update(key, forward, backward)

        self.update_input_groups()
        self.input_group_combiner.update()

        # Update own status
        valid = True
        messages = []

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('[{}] Input {} is not valid: {}'.format(self.id, input_.id, message))

        for input_group in self.input_groups.values():
            if input_group.empty:
                valid = False
                messages.append('[{}] InputGroup {} is empty'.format(self.id, input_group.id))

        for id_, output in self.outputs.items():
            if output.resulting_datatype is not None and not issubclass(output.resulting_datatype, DataType):
                valid = False
                messages.append('[{}] Output {} cannot determine the Output DataType (got {}), please specify a '
                                'valid DataType or add casts to the Links'.format(self.id,
                                                                                  id_,
                                                                                  output.resulting_datatype))

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update all outputs
        for output in self.outputs.values():
            output.update(key, forward, backward)

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, False)



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import re
import shlex
import subprocess
import sys
import threading
import time

import fastr
from fastr import exceptions
from fastr.plugins.executionplugin import ExecutionPlugin
from fastr.execution.job import JobState
from fastr.helpers.classproperty import classproperty

SBATCH_SCRIPT_TEMPLATE = \
"""#!/bin/bash
#SBATCH --nodes=1
#SBATCH --ntasks=1
#SBATCH --cpus-per-task {cores}
{memory}
{time}
{partition}
{gpus}
#SBATCH --output {stdout}
#SBATCH --error {stderr}
{depends}
{hold}

{command}
"""


class SlurmExecution(ExecutionPlugin):
    """
    The SlurmExecution plugin allows you to send the jobs to SLURM using
    the sbatch command. It is pure python and uses the sbatch, scancel,
    squeue and scontrol programs to control the SLURM scheduler.
    """

    SQUEUE_FORMAT = '{"id": %.18i, "status": "%.2t"}'

    # Job state, compact form:
    #   PD (pending),
    #    R (running),
    #   CA (cancelled),
    #   CF(configuring),
    #   CG (completing),
    #   CD (completed),
    #    F (failed),
    #   TO (timeout),
    #   NF (node failure),
    #   RV (revoked) and
    #   SE (special exit state).
    STATUS_MAPPING = {
        'PD': JobState.queued,
        ' R': JobState.running,
        'CA': JobState.cancelled,
        'CF': JobState.running,
        'CG': JobState.running,
        'CD': JobState.finished,
        ' F': JobState.failed,
        'TO': JobState.queued,
        'NF': JobState.failed,
        'RV': JobState.cancelled,
        'SE': JobState.failed,
    }

    # SLURM COMMANDS
    SBATCH = 'sbatch'
    SCANCEL = 'scancel'
    SQUEUE = 'squeue'
    SCONTROL = 'scontrol'

    # Advertise capabilities
    SUPPORTS_DEPENDENCY = True
    SUPPORTS_CANCEL = True
    SUPPORTS_HOLD_RELEASE = True

    def __init__(self, finished_callback=None, cancelled_callback=None):
        super(SlurmExecution, self).__init__(finished_callback, cancelled_callback)

        # Config value
        self.check_interval = fastr.config.slurm_job_check_interval
        self.partition = fastr.config.slurm_partition

        # Create job translation table
        self.job_translation_table = dict()
        self.job_lookup_table = dict()

        # Start job status checker
        fastr.log.debug('Creating job status checker')
        self.job_checker = threading.Thread(name='SlurmJobChecker-0', target=self.job_status_check, args=())
        self.job_checker.daemon = True
        fastr.log.debug('Starting job status checker')
        self.job_checker.start()

    def cleanup(self):
        # Stop thread
        super(SlurmExecution, self).cleanup()

        if self.job_checker.is_alive():
            fastr.log.debug('Terminating job_checker thread')
            self.job_checker.join()

    @classproperty
    def configuration_fields(cls):
        return {
            "slurm_job_check_interval": (int, 30, "The interval in which the job checker will start"
                                                  "to check for stale jobs"),
            "slurm_partition": (str, '', "The slurm partition to use")
        }

    @classmethod
    def test(cls):
        # Check if requirement commands can be called
        try:
            subprocess.check_output([cls.SBATCH, '--help'], stderr=subprocess.STDOUT)
        except OSError:
            raise exceptions.FastrExecutableNotFoundError(cls.SBATCH)

        try:
            subprocess.check_output([cls.SQUEUE, '--help'], stderr=subprocess.STDOUT)
        except OSError:
            raise exceptions.FastrExecutableNotFoundError(cls.SQUEUE)

        try:
            subprocess.check_output([cls.SCONTROL, '--help'], stderr=subprocess.STDOUT)
        except OSError:
            raise exceptions.FastrExecutableNotFoundError(cls.SCONTROL)

        try:
            subprocess.check_output([cls.SCANCEL, '--help'], stderr=subprocess.STDOUT)
        except OSError:
            raise exceptions.FastrExecutableNotFoundError(cls.SCANCEL)

    def _job_finished(self, result):
        pass

    def _queue_job(self, job):
        # Encode command to execute
        command = [sys.executable,
                   os.path.join(fastr.config.executionscript),
                   str(job.commandfile)]
        command = ' '.join(shlex.quote(item) for item in command)

        # Encode job dependencies
        depends = ''
        if job.hold_jobs:
            depend_jobs = [self.job_lookup_table.get(j, None) for j in job.hold_jobs]
            depend_jobs = [str(x) for x in depend_jobs if x is not None]
            if depend_jobs:
                depends = '#SBATCH --dependency=afterok:{}'.format(','.join(depend_jobs))

        if job.resources.memory:
            required_memory = '#SBATCH --mem {}M'.format(job.resources.memory)
        else:
            required_memory = ''

        if job.resources.gpus:
            required_gpus = '#SBATCH --gres=gpu:{}'.format(job.resources.gpus)
        else:
            required_gpus = ''

        if job.resources.time:
            hours = job.resources.time // 3600
            minutes = (job.resources.time % 3600) // 60
            seconds = job.resources.time % 60

            walltime = '{}:{:02d}:{:02d}'.format(hours, minutes, seconds)
            required_time = '#SBATCH --time {}'.format(walltime)
        else:
            required_time = ''

        if self.partition:
            partition = '#SBATCH --partition {}'.format(self.partition)
        else:
            partition = ''

        # Check if the job is ready to run or must be held
        if job.status == JobState.hold:
            hold = '--hold'
        else:
            hold = ''

        # Create SBATCH script from template
        sbatch_script = SBATCH_SCRIPT_TEMPLATE.format(
            cores=job.resources.cores or 1,
            memory=required_memory,
            gpus=required_gpus,
            time=required_time,
            partition=partition,
            stdout=job.stdoutfile,
            stderr=job.stderrfile,
            command=command,
            depends=depends,
            hold=hold,
        ).encode()

        if fastr.config.debug:
            fastr.log.debug('USING SBATCH SCRIPT:\n{}'.format(sbatch_script))

        sbatch = subprocess.Popen(
            [self.SBATCH],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )

        # Feed script via stdin and get result
        stdout, stderr = sbatch.communicate(sbatch_script)
        stdout = stdout.decode()

        match = re.search(r'Submitted batch job (\d+)', stdout)

        if match:
            cl_job_id = int(match.group(1))
        else:
            # Set errors and process job as failed
            fastr.log.error('Could not submit job, sbatch returned:\n{}\n{}'.format(stdout, stderr))
            job.status = JobState.execution_failed
            self.job_finished(job, errors=["FastrSlurmSubmitError",
                                           stderr,
                                           __file__,
                                           "unknown"])
            return

        self.job_translation_table[cl_job_id] = job
        self.job_lookup_table[job.id] = cl_job_id

    def _cancel_job(self, job):
        """
        Cancel a given job

        :param job:
        """
        try:
            slurm_job_id = self.job_lookup_table.pop(job.id)
        except KeyError:
            fastr.log.info('Job {} not found in SLURM lookup'.format(job.id))
            return

        scancel_process = subprocess.Popen([self.SCANCEL, str(slurm_job_id)])
        stdout, stderr = scancel_process.communicate()

        if stderr:
            fastr.log.warning('Encountered error when cancelling job: {}'.format(job.id))

        try:
            del self.job_translation_table[slurm_job_id]
        except KeyError:
            pass  # This job is already gone

    def _hold_job(self, job):
        slurm_job_id = self.job_lookup_table.get(job.id, None)
        if slurm_job_id:
            scontrol_process = subprocess.Popen([self.SCONTROL, "hold", str(slurm_job_id)])
            stdout, stderr = scontrol_process.communicate()
            if stderr:
                fastr.log.warning('Encountered error when holding job: {}'.format(job.id))
        else:
            fastr.log.error('Cannot hold job {}, cannot find the slurm id!'.format(job.id))

    def _release_job(self, job):
        slurm_job_id = self.job_lookup_table.get(job.id, None)
        if slurm_job_id:
            scontrol_process = subprocess.Popen([self.SCONTROL, "release", str(slurm_job_id)])
            stdout, stderr = scontrol_process.communicate()
            if stderr:
                fastr.log.warning('Encountered error when releasing job: {}'.format(job.id))
        else:
            fastr.log.error('Cannot release job {}, cannot find the slurm id!'.format(job.id))

    def job_status_check(self):
        last_update = time.time()

        while self.running:
            if time.time() - last_update < self.check_interval:
                time.sleep(1)
                continue

            fastr.log.debug('Running job status check')
            last_update = time.time()  # Reset timer
            cluster_job_ids = set(self.job_lookup_table.values()) | set(self.job_translation_table.keys())

            command = [self.SQUEUE,
                       '-o', self.SQUEUE_FORMAT,
                       '-j', ','.join(str(x) for x in cluster_job_ids)]

            squeue = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = squeue.communicate()
            stdout = stdout.decode()

            job_info = {}
            for line in stdout.splitlines()[1:]:
                if line.strip() == '':
                    continue
                job = json.loads(line)
                job_info[job['id']] = job['status']

            for job_id in cluster_job_ids:
                if job_id not in job_info:
                    # Job is done, remove from plugin tracking and dispatch callbacks
                    job = self.job_translation_table.pop(job_id, None)

                    if job is not None:
                        # Send the result to the callback function
                        try:
                            del self.job_lookup_table[job.id]
                        except KeyError:
                            fastr.log.warning('Found an inconsistency in the job_lookup_table,'
                                              ' cannot find job to remove')

                    self.job_finished(job)
                else:
                    # Update the job status
                    job = self.job_translation_table[job_id]
                    status = self.STATUS_MAPPING.get(job_info[job_id], None)

                    if status is None:
                        fastr.log.info('Found a job with unknown state: {}'.format(job_info[job_id]))
                        continue

                    job.status = status



#!/usr/bin/env python

# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def create_network():
    # Import the faster environment and set it up
    import fastr
    # Create a new network
    network = fastr.create_network(id='advanced_linking')

    # Create a source node in the network
    source1 = network.create_source('Int', id='source',
                                    resources=fastr.api.ResourceLimit(cores=1,
                                                                      memory='2G'))

    constant1 = network.create_constant('Int', data=[1, 2, 3, 4, 5], id='constant',
                                        resources=fastr.api.ResourceLimit(cores=1,
                                                                          memory='2G'))

    # Create a new node in the network using tools
    node1 = network.create_node('fastr/math/AddInt:1.0', tool_version="1.0", id="add1")
    node2 = network.create_node('fastr/math/AddInt:1.0', tool_version="1.0", id="add2")

    # Create a link between the source output and an input of the addint node
    # Link multiple outputs to one input
    link1, link2 = (source1.output, constant1.output) >> node1.inputs['left_hand']

    # Test mixed link
    links = (source1.output, 10, 11, constant1.output, 6, 7, 8) >> node2.inputs['left_hand']

    # Create a constant node and link it to the value2 input of the addint node
    node1.inputs['right_hand'] = (10, 11)
    node2.inputs['right_hand'] = (25, 26, 27, 28, 29, 30, 31)

    # Create a sink to save the data
    sink1 = network.create_sink('Int', id='sink1')
    sink2 = network.create_sink('Int', id='sink2')

    # Link the addint node to the sink
    sink1.input << node1.outputs['result']
    sink2.input << node2.outputs['result']

    return network


def source_data(network):
    return {'source': [1, 'vfslist://example_data/add_ints/values']}


def sink_data(network):
    return {
        'sink1': 'vfs://tmp/results/{}/result1_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
        'sink2': 'vfs://tmp/results/{}/result2_{{sample_id}}_{{cardinality}}{{ext}}'.format(network.id),
    }


def main():
    network = create_network()

    # Execute
    network.draw(network.id, draw_dimensions=True)
    network.execute(source_data(network), sink_data(network))


if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import json
import sys
import urllib.parse as up

import fastr
from fastr import exceptions, resources
from fastr.core.ioplugin import IOPlugin
from fastr.data import url as urltools
from fastr.utils.cmd import add_parser_doc_link


def get_parser():
    parser = argparse.ArgumentParser(description="Executes an source command",
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-i', '--input', nargs='+', type=str, required=True,
                        help="The url to process (can also be a list)")
    parser.add_argument('-o', '--output', type=str, metavar='OUTPUT', required=True,
                        help="The output url in vfs scheme")
    parser.add_argument('-d', '--datatype', default=None, type=str,
                        help="The datatype of the source/sink data to handle")
    parser.add_argument('-s', '--sample_id', default=None, type=str,
                        help="The sample_id of the source/sink data to handle")
    return parser


def source():
    # No arguments were parsed yet, parse them now
    parser = add_parser_doc_link(get_parser(), __file__)
    args = parser.parse_args()

    fastr.log.info('Using Python version {} loaded from {}'.format(sys.version, sys.executable))
    fastr.log.info(('Arguments:\ninput: {}\noutput: {}\n'
                    'datatype: {}\nsample_id: {}').format(args.input,
                                                          args.output,
                                                          args.datatype,
                                                          args.sample_id))

    inputs = args.input
    output = args.output
    sample_id = args.sample_id
    datatype = args.datatype

    if datatype is not None:
        datatype = fastr.types[datatype]
    fastr.log.debug("datatype to use: {}".format(datatype))

    if sample_id is None:
        message = 'For Source behaviour the sample id needs to be defined!'
        fastr.log.critical(message)
        raise exceptions.FastrValueError(message)

    if urltools.isurl(output):
        message = 'For Source behaviour the output should be a directory, not a URL! (found {})'.format(output)
        fastr.log.critical(message)
        raise exceptions.FastrValueError(message)

    results = []
    for input_ in inputs:
        if urltools.isurl(input_):
            parsed_input = up.urlparse(input_)
            try:
                plugin = resources.ioplugins[parsed_input.scheme]
            except KeyError:
                message = "No valid scheme is supplied in: {} (Found {}). The following schemes are supported: {}".format(
                    input_,
                    parsed_input.scheme,
                    ' '.join(resources.ioplugins.keys())
                )
                fastr.log.error(message)
                raise exceptions.FastrUnknownURLSchemeError(message)

            results.append(plugin.pull_source_data(input_, output, sample_id, datatype))
        else:
            results.append(input_)

    if len(results) == 1:
        # If we have only one result, there is no need to merge
        result = results[0]
    else:
        # Merge multiple cardinality into a single result
        samples = set()
        for part in results:
            # Check if we got value or if the value was a simple value
            if isinstance(part, dict):
                samples.update(part.keys())

        # For each sample collect all found values
        result = {}
        for sample_id in samples:
            result[sample_id] = temp = []

            for part in results:
                if isinstance(part, dict):
                    # We have a dictionary with {sampleid: (value1, value2)} format
                    temp.extend(part.get(sample_id, []))
                else:
                    # We have a value (not URL) that was given, we assume this to be used for ALL samples (broadcasting)
                    temp.append(part)

    IOPlugin.print_result(result)

    # Cleanup all plugins
    resources.ioplugins.cleanup()


def main():
    """
    Command line access to the IOPlugin source
    """
    try:
        source()
    except Exception as exception:
        # Pass error on to main process and raise error again
        exception = (
            type(exception).__name__,
            str(exception),
            exception.filename if hasattr(exception, 'filename') else None,
            exception.linenumber if hasattr(exception, 'linenumber') else None,
        )

        print('__FASTR_ERRORS__ = {}'.format(
            json.dumps(
                [exception]
            )
        ))
        raise

if __name__ == '__main__':
    main()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The module containing the classes describing the targets.
"""

import os
import platform
import subprocess
from typing import List

import fastr
from fastr import exceptions
from fastr.core.target import SubprocessBasedTarget, TargetResult
from fastr.data import url

# Check if environment modules are available
try:
    from fastr.execution.environmentmodules import EnvironmentModules
    ENVIRONMENT_MODULES = EnvironmentModules(fastr.config.protected_modules)
    ENVIRONMENT_MODULES_LOADED = True
except exceptions.FastrValueError:
    ENVIRONMENT_MODULES = None
    ENVIRONMENT_MODULES_LOADED = False


class LocalBinaryTarget(SubprocessBasedTarget):
    """
    A tool target that is a local binary on the system. Can be found using
    environmentmodules or a path on the executing machine. A local binary
    target has a number of fields that can be supplied:

    * ``binary (required)``: the name of the binary/script to call, can also be called ``bin``
      for backwards compatibility.
    * ``modules``: list of modules to load, this can be environmentmodules or lmod
      modules. If modules are given, the ``paths``, ``environment_variables`` and ``initscripts``
      are ignored.
    * ``paths``: a list of paths to add following the structure ``{"value": "/path/to/dir", "type": "bin"}``.
      The types can be ``bin`` if the it should be added to $PATH or ``lib`` if it should be
      added to te library path (e.g. $LD_LIBRARY_PATH for linux).
    * ``environment_variables``: a dictionary of environment variables to set.
    * ``initscript``: a list of script to run before running the main tool
    * ``interpreter``: the interpreter to use to call the binary e.g. ``python``

    The LocalBinaryTarget will first check if there are modules given and the module subsystem is loaded.
    If that is the case it will simply unload all current modules and load the given modules. If not it
    will try to set up the environment itself by using the following steps:

    1. Prepend the bin paths to $PATH
    2. Prepend the lib paths to the correct environment variable
    3. Setting the other environment variables given ($PATH and the system
       library path are ignored and cannot be set that way)
    4. Call the initscripts one by one

    The definition of the target in JSON is very straightforward:

    .. code-block:: json

        {
          "binary": "bin/test.py",
          "interpreter": "python",
          "paths": [
            {
              "type": "bin",
              "value": "vfs://apps/test/bin"
            },
            {
              "type": "lib",
              "value": "./lib"
            }
          ],
          "environment_variables": {
            "othervar": 42,
            "short_var": 1,
            "testvar": "value1"
          },
          "initscripts": [
            "bin/init.sh"
          ],
          "modules": ["elastix/4.8"]
        }

    In XML the definition would be in the form of:

   .. code-block:: xml

        <target os="linux" arch="*" modules="elastix/4.8" bin="bin/test.py" interpreter="python">
          <paths>
            <path type="bin" value="vfs://apps/test/bin" />
            <path type="lib" value="./lib" />
          </paths>
          <environment_variables short_var="1">
            <testvar>value1</testvar>
            <othervar>42</othervar>
          </environment_variables>
          <initscripts>
            <initscript>bin/init.sh</initscript>
          </initscripts>
        </target>
    """

    DYNAMIC_LIBRARY_PATH_DICT = {
        'windows': 'PATH',  # Not Tested
        'linux': 'LD_LIBRARY_PATH',  # Tested
        'darwin': 'DYLD_LIBRARY_PATH',  # Tested
    }

    _platform = platform.system().lower()
    if _platform not in DYNAMIC_LIBRARY_PATH_DICT:
        fastr.log.warning('"Dynamic library path not supported on platform: {}"'.format(_platform))

    def __init__(self, binary, paths=None, environment_variables=None,
                 initscripts=None, modules=None, interpreter=None, **kwargs):
        """
        Define a new local binary target. Must be defined either using paths and optionally environment_variables
        and initscripts, or enviroment modules.
        """
        if binary.startswith('./'):
            binary = os.path.abspath(binary)

        self.binary = binary
        if modules is None:
            if 'module' in kwargs and kwargs['module'] is not None:
                fastr.log.warning('Using deprecated module in target (modules is new way to do it)')
                self._modules = (kwargs['module'],)
            else:
                self._modules = None
        elif isinstance(modules, str):
            self._modules = (modules.strip(),)
        else:
            self._modules = tuple(x.strip() for x in modules)

        if isinstance(paths, str):
            self._paths = [{'type': 'bin', 'value': paths}]
        elif paths is None and 'location' in kwargs and kwargs['location'] is not None:
            fastr.log.warning('Using deprecated location in target (paths is the new way to do it)')
            self._paths = [{'type': 'bin', 'value': kwargs['location']}]
        else:
            self._paths = paths

        if self._paths is not None:
            for path_entry in self._paths:
                if not url.isurl(path_entry['value']):
                    fastr.log.info('Changing {}'.format(path_entry['value']))
                    path_entry['value'] = os.path.abspath(path_entry['value'])

        if environment_variables is None:
            environment_variables = {}
        self._envvar = environment_variables

        if initscripts is None:
            initscripts = []
        self._init_scripts = [os.path.abspath(x) for x in initscripts]

        self.interpreter = interpreter

        self._roll_back = None

    def __enter__(self):
        """
        Set the environment in such a way that the target will be on the path.
        """
        super(LocalBinaryTarget, self).__enter__()

        # Create dictionary of possible platforms, to set dynamic labrary path
        # Add check to see if _platform is present in dictionary
        if self._platform in self.DYNAMIC_LIBRARY_PATH_DICT:
            dynamic_library_path = self.DYNAMIC_LIBRARY_PATH_DICT[self._platform]
        else:
            dynamic_library_path = None

        if ENVIRONMENT_MODULES_LOADED and self._modules is not None and len(self._modules) > 0:
            # Clear the enviroment modules and load all required modules
            ENVIRONMENT_MODULES.clear()
            for module_ in self._modules:
                if not ENVIRONMENT_MODULES.isloaded(module_):
                    ENVIRONMENT_MODULES.load(module_)
                    fastr.log.info('loaded module: {}'.format(module_))

                if not ENVIRONMENT_MODULES.isloaded(module_):
                    raise exceptions.FastrImportError('EnvironmentModule {} could not be loaded!'.format(module_))
            fastr.log.info('LoadedModules: {}'.format(ENVIRONMENT_MODULES.loaded_modules))
        elif self._paths is not None:
            # Prepend PATH and LD_LIBRARY_PATH as required
            self._roll_back = {'PATH': os.environ.get('PATH', None)}

            # Prepend extra paths to PATH
            bin_path = os.environ.get('PATH', None)
            bin_path = [bin_path] if bin_path else []
            extra_path = [x['value'] for x in self._paths if x['type'] == 'bin']
            extra_path = [fastr.vfs.url_to_path(x) if url.isurl(x) else x for x in extra_path]
            fastr.log.info('Adding extra PATH: {}'.format(extra_path))
            os.environ['PATH'] = os.pathsep.join(extra_path + bin_path)

            # Prepend extra paths to LB_LIBRARY_PATH
            extra_ld_library_path = [x['value'] for x in self._paths if x['type'] == 'lib']
            if len(extra_ld_library_path) > 0:
                if dynamic_library_path is None:
                    message = 'Cannot set dynamic library path on platform: {}'.format(self._platform)
                    fastr.log.critical(message)
                    raise exceptions.FastrNotImplementedError(message)

                self._roll_back[dynamic_library_path] = os.environ.get(dynamic_library_path, None)

                lib_path = os.environ.get(dynamic_library_path, None)
                lib_path = [lib_path] if lib_path else []
                extra_ld_library_path = [fastr.vfs.url_to_path(x) if url.isurl(x) else x for x in extra_ld_library_path]

                fastr.log.info('Adding extra LIB: {}'.format(extra_ld_library_path))
                os.environ[dynamic_library_path] = os.pathsep.join(extra_ld_library_path + lib_path)

            # Set other environment variables as indicated
            for var, value in self._envvar.items():
                if var in ['PATH', dynamic_library_path]:
                    continue

                self._roll_back[var] = os.environ.get(var, None)
                os.environ[var] = str(value)

            # Run init script(s) if required
            for script in self._init_scripts:
                if isinstance(script, str):
                    script = [script]

                subprocess.call(script)
        else:
            raise exceptions.FastrNotImplementedError(
                'Binary targets must have either paths or modules set! (binary {})'.format(self.binary)
            )

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Cleanup the environment
        """
        if ENVIRONMENT_MODULES_LOADED and self._modules is not None and len(self._modules) > 0:
            ENVIRONMENT_MODULES.clear()
        elif self._roll_back is not None:
            for var, value in self._roll_back.items():
                if value is not None:
                    os.environ[var] = value
                else:
                    del os.environ[var]

            self._roll_back = None

    def run_command(self, command: List) -> TargetResult:
        if self.interpreter is not None:
            paths = [x['value'] for x in self._paths if x['type'] == 'bin']
            fastr.log.info('Options: {}'.format(paths))
            try:
                containing_path = next(x for x in paths if os.path.exists(os.path.join(x, command[0])))
            except StopIteration:
                raise exceptions.FastrScriptNotFoundError(
                    self.interpreter,
                    command[0],
                    paths
                )
            interpreter = self.interpreter

            command = [interpreter, os.path.join(containing_path, command[0])] + command[1:]

        fastr.log.debug('COMMAND: "{}" ({})'.format(command, type(command).__name__))
        return self.call_subprocess(command)

    @property
    def paths(self):
        return self._paths



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imp
import os
import shutil

import fastr
from fastr.examples.test.test_examples import EXAMPLE_LIST


def create_reference(example):
    examples_dir = fastr.config.examplesdir
    example_path = os.path.abspath(os.path.join(examples_dir, example + '.py'))
    example_module = imp.load_source(example, example_path)

    network = example_module.create_network()
    output_directory = './data/reference/{n.id}_{n.version}'.format(n=network)

    if os.path.exists(output_directory):
        fastr.log.info('Skipping creation of {}, already exists!'.format(output_directory))
        return

    # Do not publish this to PIM
    fastr.config.pim_host = ''  # Do not publish anything to PIM

    source_data = example_module.source_data(network)
    network.create_reference(source_data=source_data, output_directory=output_directory)
    shutil.rmtree(os.path.join(output_directory, '__fastr_run_tmp__'))


def generate_all_references():
    for example in EXAMPLE_LIST:
        create_reference(example)


if __name__ == '__main__':
    generate_all_references()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the virtual file system code. This is internally
used object as used as base class for the IOPlugin.
"""

from fastr.core.vfs import VirtualFileSystem
from fastr.core.ioplugin import IOPlugin


class VirtualFileSystem(VirtualFileSystem, IOPlugin):
    """
    The virtual file system class. This is an IOPlugin, but also heavily used
    internally in fastr for working with directories. The VirtualFileSystem
    uses the ``vfs://`` url scheme.

    A typical virtual filesystem url is formatted as ``vfs://mountpoint/relative/dir/from/mount.ext``

    Where the ``mountpoint`` is defined in the :ref:`config-file`. A list of
    the currently known mountpoints can be found in the ``fastr.config`` object

    .. code-block:: python

        >>> fastr.config.mounts
        {'example_data': '/home/username/fastr-feature-documentation/fastr/fastr/examples/data',
         'home': '/home/username/',
         'tmp': '/home/username/FastrTemp'}

    This shows that a url with the mount ``home`` such as
    ``vfs://home/tempdir/testfile.txt`` would be translated into
    ``/home/username/tempdir/testfile.txt``.

    There are a few default mount points defined by Fastr (that can be changed
    via the config file).

    +--------------+-----------------------------------------------------------------------------+
    | mountpoint   | default location                                                            |
    +==============+=============================================================================+
    | home         | the users home directory (:py:func:`expanduser('~/') <os.path.expanduser>`) |
    +--------------+-----------------------------------------------------------------------------+
    | tmp          | the fastr temprorary dir, defaults to ``tempfile.gettempdir()``             |
    +--------------+-----------------------------------------------------------------------------+
    | example_data | the fastr example data directory, defaults ``$FASTRDIR/example/data``       |
    +--------------+-----------------------------------------------------------------------------+

    """
    scheme = 'vfs'



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import gzip
import os

from fastr.datatypes import URLType


class NiftiImageFileCompressed(URLType):
    description = 'Compressed Nifti Image File format'
    extension = 'nii.gz'

    def _validate(self):
        parsed_value = self.parsed_value

        if self.extension and not parsed_value.endswith(self.extension):
            return False

        if not os.path.isfile(parsed_value):
            return False

        # Check magic bytes for Nifti-1 and Nifti-2 headers
        # type depends on header size and the both have magic bits for
        # identification
        try:
            with gzip.open(parsed_value, 'rb') as fin:
                header_size_bytes = fin.read(4)

                if len(header_size_bytes) != 4:
                    return False

                header_size = int.from_bytes(header_size_bytes, 'little')
                if header_size not in [348, 540]:
                    header_size = int.from_bytes(header_size_bytes, 'big')

                if header_size == 348:
                    fin.seek(344)
                    magic = fin.read(4)
                    return magic == b'n+1\x00'
                elif header_size == 540:
                    magic = fin.read(8)
                    return magic == b'\x6E\x2B\x32\x00\x0D\x0A\x1A\x0A'
                else:
                    return False
        except OSError:
            # Not a valid gzip archive
            return False

    def checksum(self):
        """
        Return the checksum of this URL type

        :return: checksum string
        :rtype: str
        """
        contents = self.content(self.parsed_value)
        hasher = hashlib.new('md5')

        for path in contents:
            with gzip.open(path, 'rb') as file_handle:
                while True:
                    data = file_handle.read(32768)
                    if not data:
                        break
                    hasher.update(data)
        return hasher.hexdigest()



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import imp

import pytest

import fastr

EXAMPLE_LIST = [
    'source_sink',
    'collapse',
    'expand',
    'collapse_expand',
    'filecopy',
    'add_ints',
    'shift_links',
    'auto_prefix',
    'chunk_test',
    'input_groups',
    'failing_network',
    'advanced_linking',
    'macro_node',
    'macro_node2',
    'failing_macro',
]


@pytest.mark.parametrize("example_name", EXAMPLE_LIST)
@pytest.mark.slow
def test_examples(example_name, tmpdir_factory):
    # Set some config variables
    example_path = os.path.join(fastr.config.examplesdir, example_name + '.py')
    example_module = imp.load_source(example_name, example_path)
    network = example_module.create_network().parent

    reference_dir = os.path.join(os.path.dirname(example_path),
                                 'data',
                                 'reference',
                                 '{}_{}'.format(network.id, network.version))

    fastr.log.info('Using reference dir: {}'.format(reference_dir))
    if os.path.isdir(reference_dir):
        result = network.test(reference_dir, network)
        if result:
            for error in result:
                print(error)
        assert(len(result) == 0)
    else:
        raise ValueError('Could not find reference data for test!')




# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A module to maintain a network node.

Exported classes:

Node -- A class encapsulating a tool.
ConstantNode -- A node encapsulating an Output to set scalar values.
SourceNode -- A class providing a handle to a file.
"""
import itertools
import os
import pprint
import weakref
from abc import ABCMeta
from collections import OrderedDict, defaultdict

import sympy

from .. import api
from .. import exceptions
from ..abc.serializable import Serializable
from ..abc.updateable import Updateable
from ..core.dimension import HasDimensions, Dimension
from ..core.resourcelimit import ResourceLimit
from ..core.interface import InputSpec, OutputSpec
from ..core.samples import SampleId
from ..core.tool import Tool
from ..datatypes import DataType, types
from ..helpers import log
from ..planning.inputoutput import BaseInput, Input, MacroInput, Output, AdvancedFlowOutput, SourceOutput, MacroOutput
from ..planning.inputgroup import InputGroup
from ..planning.inputgroupcombiner import DefaultInputGroupCombiner, MergingInputGroupCombiner
from .. import resources


class InputDict(OrderedDict):
    """
    The container containing the Inputs of Node. Implements helper functions
    for the easy linking syntax.
    """

    # We know this class is not really for public interaction, however it has
    # an important function linking of nodes.
    # pylint: disable=too-few-public-methods

    def __setitem__(self, key, value):
        """
        Set an item in the input dictionary. The behaviour depends on the
        type of the value. For a
        :py:class:`BaseInput <fastr.planning.inputoutput.BaseInput>`,
        the input will simply be added to the list of inputs. For a
        :py:class:`BaseOutput <fastr.planning.inputoutput.BaseOutput>`,
        a link between the output and input will be created.

        :param str key: id of the input to assign/link
        :param value: either the input to add or the output to link
        :type value: :py:class:`BaseInput <fastr.planning.inputoutput.BaseInput>`
                      or
                      :py:class:`BaseOutput <fastr.planning.inputoutput.BaseOutput>`
        """
        if isinstance(value, Input):
            super(InputDict, self).__setitem__(key, value)
        else:
            self[key].create_link_from(value)


class OutputDict(OrderedDict):
    """
    The container containing the Inputs of Node. Only checks if the inserted
    values are actually outputs.
    """

    # We know this class is not really for public interaction, however it we
    # have it to do type checking and consistency with the InputDict
    # pylint: disable=too-few-public-methods

    def __setitem__(self, key, value):
        """
        Set an output.

        :param str key: the of the item to set
        :param value: the output to set
        :type value: :py:class:`BaseOutput <fastr.planning.inputoutput.BaseOutput>`
        """
        if isinstance(value, Output):
            super(OutputDict, self).__setitem__(key, value)
        else:
            message = 'Cannot add object of type {} to OutputDict'.format(type(value).__name__)
            log.warning(message)


class BaseNode(HasDimensions, Updateable, Serializable, metaclass=ABCMeta):
    NODE_TYPES = {}

    def __init_subclass__(cls, **kwargs):
        """
        Register nodes in class for easly location
        """
        super().__init_subclass__(**kwargs)
        cls.NODE_TYPES[cls.__name__] = cls


class Node(BaseNode):
    """
    The class encapsulating a node in the network. The node is responsible
    for setting and checking inputs and outputs based on the description
    provided by a tool instance.
    """
    __dataschemafile__ = 'Node.schema.json'

    _InputType = Input
    _OutputType = Output

    def __init__(self, tool, id_=None, node_class=None, parent=None, resource_limits=None, nodegroup=None):
        """
        Instantiate a node.

        :param tool: The tool to base the node on
        :type tool: :py:class:`Tool <fastr.core.tool.Tool>`
        :param str id_: the id of the node
        :param str node_class: The class of the NodeRun to create (e.g. SourceNodeRun, NodeRun)
        :param parent: the parent network of the node
        :type parent: :py:class:`Network <fastr.planning.network.Network>`
        :return: the newly created Node
        """
        super(Node, self).__init__()

        if isinstance(tool, Tool):
            self._tool = tool
        elif isinstance(tool, (str, tuple)):
            if tool in resources.tools:
                self._tool = resources.tools[tool]
            else:
                message = ('Specified tool ({}) is not in the tools: {}. '
                           'Check the config (fastr/resources/fastr.config)').format(tool,
                                                                                     list(resources.tools.todict().keys()))
                log.critical(message)
                raise exceptions.FastrToolUnknownError(message)
        elif tool is None:
            self._tool = None
        else:
            message = 'tool should either be a string or a Tool.'
            log.critical(message)
            raise exceptions.FastrTypeError(message)

        # Don't set parent here, as not info needed for registration is there yet
        self._parent = None
        if parent is None:
            message = 'parent argument is None, need a parent Network to function!'
            raise exceptions.FastrValueError(message)

        node_number = parent.node_number
        parent.node_number += 1

        if id_ is None:
            # Node ID is a simple $name_$counter format, making sure nodes can
            # not be named the same

            #: The Node id s a unique string identifying the Node
            id_ = 'node_{}_{}'.format(self.name, node_number)

        parent.check_id(id_)
        self._id = id_
        self._nodegroup = nodegroup

        #: The parent is the Network this Node is part of
        self.parent = parent

        #: A list of inputs of this Node
        self.inputs = InputDict()

        #: A list of outputs of this Node
        self.outputs = OutputDict()

        # Create all inputs and outputs, if the class is set in the Tool file,
        # respect that, otherwise use the class default.
        if self._tool is not None:
            for name, input_ in self._tool.inputs.items():
                self.inputs[name] = self._InputType(self, input_)
            for name, output in self._tool.outputs.items():
                self.outputs[name] = self._OutputType(self, output)

        # Set the job requirements
        self.resources = resource_limits

        # Cache dimensions for speed
        self._dimensions = None

        # Set the flow control
        self._input_groups = OrderedDict()
        self._merge_dimensions = None
        self.input_group_combiner = None
        self.merge_dimensions = 'none'

        # Update Inputs and self (which calls Outputs)
        self.update()

    def __repr__(self):
        """
        Get a string representation for the Node

        :return: the string representation
        :rtype: str
        """
        if self._tool is not None:
            toolinfo = '(tool: {tool.id} v{tool.version!s})'.format(tool=self._tool)
        else:
            toolinfo = ''
        return_list = ['{} {} {}'.format(type(self).__name__, self.id, toolinfo)]

        # The "+ [8]" guarantees a minimum of 8 width and avoids empty lists
        width_input_keys = max([len(x) for x in self.inputs.keys()] + [8])
        width_input_types = max([len(x.datatype.id) for x in self.inputs.values()] + [8]) + 2
        width_output_keys = max([len(x) for x in self.outputs.keys()] + [8])
        width_output_types = max([len(x.datatype.id) for x in self.outputs.values()] + [8]) + 2

        return_list.append('{:^{}}  | {:^{}}'.format('Inputs', width_input_types + width_input_keys + 1,
                                                     'Outputs', width_output_types + width_output_keys + 1))
        return_list.append('-' * (width_input_keys + width_input_types + width_output_keys + width_output_types + 7))
        for (input_key, input_, output_key, output) in itertools.zip_longest(self.inputs.keys(), 
                                                                             self.inputs.values(),
                                                                             self.outputs.keys(),
                                                                             self.outputs.values()):
            if input_ is None:
                input_id = ''
                input_type = ''
            else:
                input_id = input_key
                input_type = '({})'.format(input_.datatype.id)

            if output is None:
                output_id = ''
                output_type = ''
            else:
                output_id = output_key
                output_type = '({})'.format(output.datatype.id)

            return_list.append('{:{}} {:{}}  |  {:{}} {:{}}'.format(input_id, width_input_keys,
                                                                    input_type, width_input_types,
                                                                    output_id, width_output_keys,
                                                                    output_type, width_output_types))

        return '\n'.join(return_list)

    def __str__(self):
        """
        Get a string version for the Node

        :return: the string version
        :rtype: str
        """
        return "<{}: {}>".format(type(self).__name__, self.id)

    def __eq__(self, other):
        """
        Check two Node instances for equality.

        :param other: the other instances to compare to
        :type other: fastr.planning.node.Node
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, Node):
            return NotImplemented

        if self._id != other.id:
            return False

        if self.resources != other.resources:
            return False

        if self.nodegroup != other.nodegroup:
            return False

        if self._merge_dimensions != other._merge_dimensions:
            return False

        if self.tool != other.tool:
            return False

        if self.inputs != other.inputs:
            return False

        if self.outputs != other.outputs:
            return False

        return True

    def __ne__(self, other):
        """
        Check two Node instances for inequality. This is the inverse of __eq__

        :param other: the other instances to compare to
        :type other: fastr.planning.node.Node
        :returns: True if unequal, False otherwise
        """
        if not isinstance(self, type(other)):
            return NotImplemented

        return not self == other

    def __getstate__(self):
        """
        Retrieve the state of the Node

        :return: the state of the object
        :rtype dict:
        """
        state = super(Node, self).__getstate__()

        # Make id prettier in the result
        state['id'] = self.id
        if self.nodegroup:
            state['nodegroup'] = self.nodegroup

        # Add the class of the Node in question
        state['class'] = type(self).__name__

        # Get all input and output
        state['inputs'] = [x.__getstate__() for x in self.inputs.values()]
        state['outputs'] = [x.__getstate__() for x in self.outputs.values()]

        if self._tool is not None:
            state['tool'] = {'id': self._tool.ns_id, 'version': str(self._tool.version)}

        # Add resource requirements
        state['resources'] = self.resources.__getstate__() if self.resources else self.resources
        state['merge_dimensions'] = self._merge_dimensions

        return state

    def __setstate__(self, state):
        """
        Set the state of the Node by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        # Make sure the Node classes are aligned (and warn if not so)
        if 'class' in state and state['class'] != type(self).__name__:
            log.warning('Attempting to set the state of a {} to {}'.format(
                state['class'],
                type(self).__name__
            ))

        if not hasattr(self, '_input_groups'):
            self._input_groups = OrderedDict()

        if 'id' in state:
            self._id = state['id']

        self.nodegroup = state.pop('nodegroup', None)
        parent = state.pop('parent', None)

        if 'tool' in state and state['tool'] is not None:
            if isinstance(state['tool'], list):
                tool_spec = tuple(state['tool'])
            else:
                tool_spec = (state['tool']['id'], state['tool']['version'])
            self._tool = resources.tools[tool_spec]
        else:
            self._tool = None

        # Create Input/Output objects
        inputlist = []
        for input_ in state['inputs']:
            if 'node' in input_:
                # Check if the expected Node id matches our current id
                if input_['node'] != state['id']:
                    raise exceptions.FastrParentMismatchError('This Input has different parent node!')
                del input_['node']

            # It can happen that this has been done by a subclass, check first
            if not isinstance(input_, BaseInput):
                description = self.tool.inputs[input_['id']]
                inputobj = self._InputType(self, description)
                inputobj._node = self
                inputobj.__setstate__(input_)
            else:
                inputobj = input_
            inputlist.append((inputobj.id, inputobj))

        outputlist = []
        for output in state['outputs']:
            if '_node' in output:
                # Check if the expected Node id matches our current id
                if output['_node'] != state['_id']:
                    raise exceptions.FastrParentMismatchError('This Input has different parent node!')
                del output['_node']

            # It can happen that this has been done by a subclass, check first
            if not isinstance(output, Output):
                description = self.tool.outputs[output['id']]
                outputobj = self._OutputType(self, description)
                outputobj.__setstate__(output)
            else:
                outputobj = output
            outputlist.append((outputobj.id, outputobj))

        self.inputs = InputDict(inputlist)
        self.outputs = OutputDict(outputlist)

        super(Node, self).__setstate__(state)

        if parent is not None:
            self.parent = parent
        else:
            message = 'parent argument is None, need a parent Network to function!'
            raise exceptions.FastrValueError(message)

        self._dimensions = None
        self.resources = ResourceLimit()
        if state.get('resources'):
            self.resources.__setstate__(state['resources'])

        self.merge_dimensions = state['merge_dimensions']

        self.update()

    @property
    def merge_dimensions(self):
        return self._merge_dimensions

    @merge_dimensions.setter
    def merge_dimensions(self, value):
        if isinstance(value, str):
            options = ['all', 'none']
            if value not in options:
                raise exceptions.FastrValueError('Invalid option {} given (valid options: {})'.format(value, options))
            self._merge_dimensions = value
            if value == 'none':
                self.input_group_combiner = DefaultInputGroupCombiner(self)
            elif value == 'all':
                self.input_group_combiner = MergingInputGroupCombiner(self, value)
        else:
            self._merge_dimensions = value
            self.input_group_combiner = MergingInputGroupCombiner(self, tuple(value))

    @classmethod
    def createobj(cls, state, network=None):
        if 'parent' not in state:
            if network is not None:
                log.debug('Setting network to: {}'.format(network))
                state['parent'] = network
            else:
                log.debug('No network given for de-serialization')
        else:
            log.debug('Parent is already defined as: {}'.format(network))

        state = dict(state)

        return super(Node, cls).createobj(state, network)

    @property
    def blocking(self):
        """
        Indicate that the results of this Node cannot be determined without first executing the Node, causing a
        blockage in the creation of jobs. A blocking Nodes causes the Chunk borders.
        """
        for output in self.outputs.values():
            if output.blocking:
                log.debug('Blocking because Output {} has cardinality {}'.format(output.fullid,
                                                                                       output.cardinality()))
                return True
        return False

    @property
    def dimnames(self):
        """
        Names of the dimensions in the Node output. These will be reflected
        in the SampleIdList of this Node.
        """
        if hasattr(self, '_dimnames') and self._dimnames is not None:
            return self._dimnames
        else:
            return super(Node, self).dimnames

    @dimnames.setter
    def dimnames(self, value):
        if isinstance(value, str):
            value = value,

        if not isinstance(value, tuple) or not all(isinstance(x, str) for x in value):
            raise exceptions.FastrTypeError('Dimnames has to be a tuple of str!')

        log.warning('You are overriding the dimnames of a Node, beware this is possible but not encouraged and can lead to strange results!')
        self._dimnames = value

    @dimnames.deleter
    def dimnames(self):
        del self._dimnames

    @property
    def fullid(self):
        """
        The full defining ID for the Node inside the network
        """
        return '{}/nodelist/{}'.format(self.parent.fullid, self.id)

    @property
    def global_id(self):
        """
        The global defining ID for the Node from the main network (goes out
        of macro nodes to root network)
        """
        return '{}/nodelist/{}'.format(self.parent.global_id, self.id)

    @property
    def input_groups(self):
        """
        A list of input groups for this Node. An input group is InputGroup
         object filled according to the Node

        """
        return self._input_groups

    @property
    def dimensions(self):
        if self._dimensions is None:
            self._dimensions = self.input_group_combiner.dimensions
        return self._dimensions

    @property
    def outputsize(self):
        """
        The size of output of this SourceNode
        """
        return self.size

    @property
    def nodegroup(self):
        return self._nodegroup

    @nodegroup.setter
    def nodegroup(self, value):
        self._nodegroup = value

    @property
    def id(self):
        """
        The id of the Node
        """
        return self._id

    @property
    def listeners(self):
        """
        All the listeners requesting output of this node, this means the
        listeners of all Outputs and SubOutputs
        """
        listeners = []
        for output in self.outputs.values():
            for listener in output.listeners:
                if listener not in listeners:
                    listeners.append(listener)
        return listeners

    @property
    def name(self):
        """
        Name of the Tool the Node was based on. In case a Toolless Node was
        used the class name is given.
        """
        if hasattr(self, '_tool') and isinstance(self._tool, Tool):
            return self._tool.id
        else:
            return self.__class__.__name__

    @property
    def parent(self):
        """
        The parent network of this node.
        """
        return self._parent()

    @parent.setter
    def parent(self, value):
        """
        The parent network of this node. (setter)
        """
        if hasattr(self, '_parent') and self._parent is not None:
            if self._parent() is value:
                return  # Setting to same value doesn't do anything
            raise exceptions.FastrAttributeError('Cannot reset attribute once set')

        self._parent = weakref.ref(value)
        self._parent().add_node(self)

    @property
    def status(self):
        return self._status

    @property
    def tool(self):
        return self._tool

    def draw_id(self, context):
        return '{}__{}'.format('_'.join(x.id for x in context['network_stack']), self.id)

    def draw_link_target(self, context, port_name, input=True):
        if input:
            port_spec = 'i_{}'.format(port_name)
        else:
            port_spec = 'o_{}'.format(port_name)

        return '{}:{}'.format(self.draw_id(context), port_spec)

    def draw(self, context, graph, color=None):
        if color is None:
            color = context['colors']['node']

        inputs = []
        dimensions = ''
        for input_ in self.inputs.values():
            # Skip rendering unconnected inputs
            if context['hide_unconnected']:
                if len(input_.get_sourced_nodes()) == 0:
                    continue

            # Draw dimensions in the graph
            if context['draw_dimensions']:
                dimensions = '[' + 'x'.join(context['dimensions'][x] for x in input_.dimnames) + ']'

            inputs.append('<i_{id}>{dim} {id}'.format(id=input_.id, dim=dimensions))
        inputs = '|'.join(inputs)

        outputs = []
        for output in self.outputs.values():
            # Skip rendering unconnected outputs
            if context['hide_unconnected']:
                if len(output.listeners) == 0:
                    continue
            
            # Draw dimensions in the graph
            if context['draw_dimensions']:
                dimensions = '[' + 'x'.join(context['dimensions'][x] for x in output.dimnames) + ']'

            outputs.append('<o_{id}>{id} {dim}'.format(id=output.id, dim=dimensions))
        outputs = '|'.join(outputs)

        graph.node(self.draw_id(context=context),
                   label="<id>{id}|{{{{{inputs}}}|{{{outputs}}}}}".format(id=self.id, inputs=inputs, outputs=outputs),
                   fillcolor=color,
                   style="filled")

    def get_sourced_nodes(self):
        """
        A list of all Nodes connected as sources to this Node

        :return: list of all nodes that are connected to an input of this node
        """
        sourced_node = []
        for input_ in self.inputs.values():
            for node in input_.get_sourced_nodes():
                if node not in sourced_node:
                    sourced_node.append(node)
        return sourced_node

    def find_source_index(self, target_index, target, source):
        # If there are multiple input groups, select only part of index from
        # the inputgroup which source belongs to
        if len(self.input_groups) > 1:
            input_groups = self.input_groups
            mask = [True if ig.id == source.input_group else False for ig in input_groups.values() for _ in ig.size]
            target_index = tuple(k for k, m in zip(target_index, mask) if m)

        # Delegate to InputGroup to check mixing within InputGroup
        return self.input_groups[source.input_group].find_source_index(target_size=target.size,
                                                                       target_dimnames=target.dimnames,
                                                                       source_size=source.size,
                                                                       source_dimnames=source.dimnames,
                                                                       target_index=target_index)

    def _update(self, key, forward=True, backward=False):
        """
        Update the Node information and validity of the Node and propagate
         the update downstream. Updates inputs, input groups, outputsize and outputs.

        A Node is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.planning.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized

        """
        # Make sure the Inputs and input groups are up to date
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        if backward:
            for sourced_node in self.get_sourced_nodes():
                sourced_node.update(key, False, backward)

        for input_ in self.inputs.values():
            input_.update(key, forward, backward)

        # Reset dimensions cache
        self._dimensions = None

        self.update_input_groups()
        self.input_group_combiner.update()

        # Update own status
        valid = True
        messages = []

        if len(self.get_sourced_nodes()) == 0:
            valid = False
            messages.append('[{}] No nodes are linked to any input of this Node'.format(self.id))

        if len(self.listeners) == 0:
            valid = False
            messages.append('[{}] No nodes are linked to any output of this Node'.format(self.id))

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('[{}] Input {} is not valid: {}'.format(self.id, input_.id, message))

        for input_group in self.input_groups.values():
            if input_group.empty:
                valid = False
                messages.append('[{}] InputGroup {} is empty'.format(self.id, input_group.id))

        for id_, output in self.outputs.items():
            if output.resulting_datatype is not None and not issubclass(output.resulting_datatype, DataType):
                valid = False
                messages.append('[{}] Output {} cannot determine the Output DataType (got {}), please specify a '
                                'valid DataType or add casts to the Links'.format(self.id,
                                                                                  id_,
                                                                                  output.resulting_datatype))

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update all outputs
        for output in self.outputs.values():
            output.update(key, forward, backward)

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, False)

    def update_input_groups(self):
        """
        Update all input groups in this node
        """
        input_groups = OrderedDict()

        for input_ in self.inputs.values():
            if input_.input_group not in input_groups:
                input_groups[input_.input_group] = InputGroup(self, input_.input_group)
            input_groups[input_.input_group][input_.id] = input_

        self._input_groups = input_groups


class MacroNode(Node):
    """
    MacroNode encapsulates an entire network in a single node.
    """
    _OutputType = MacroOutput
    _InputType = MacroInput

    def __init__(self, value, id_=None, parent=None, resource_limits=None, nodegroup=None):
        """
        :param value: network to create macronode for
        """
        # If MacroNode is loaded as a tool retrieve macro definition file(.py .xml .pickle .json) location
        if isinstance(value, Tool):
            if value.node_class != 'MacroNode':
                raise exceptions.FastrValueError('Tool {} is node_class is not MacroNode but {}'.format(
                    value.ns_id, value.node_class
                ))

            if not isinstance(value.target, resources.targets['MacroTarget']):
                raise exceptions.FastrTypeError('Can only use a Tool that has a Macro target to create a MacroNode')

            self._network = value.target.network
        elif isinstance(value, str) and os.path.isfile(value):
            # xml pickle, json, etc
            self._network = api.Network.load(value).parent
        # Else it is a Network class that we can use, make a copy
        else:
            try:
                # Make sure to make a copy of the network
                self._network = api.create_network_copy(value).parent
            except (ValueError, AttributeError):
                message = 'Macro node should either be a Network, a MacroTool or a FileName'
                log.critical(message)
                raise

        # This must be set before the update is called from the superclass __init__
        self._source_dimensions = {}
        self._source_sizes = {}
        self._output_info = {sink_id: None for sink_id in self._network.sinklist.keys()}

        # Now we can safely call the super
        super(MacroNode, self).__init__(None, id_, parent=parent, resource_limits=resource_limits, nodegroup=nodegroup)

        # Cache this to avoid costly recalculations
        self._create_inputs()
        self._create_outputs()

        try:
            if not self._network.is_valid():
                raise exceptions.FastrValueError('[{}] internal Network is not valid'.format(self.id))
        except exceptions.FastrError as exception:
            message = 'Macro Node: {} is not a valid network ({})'.format(id_, exception)
            log.critical(message)
            raise exceptions.FastrValueError(message)

    @property
    def network(self):
        return self._network

    def _create_inputs(self):
        for _, source in sorted(self._network.sourcelist.items()):
            spec = InputSpec(source.id, '1-*', source.datatype, required=True)
            self.inputs[spec.id] = self._InputType(self, spec)

    def _create_outputs(self):
        for _, sink in sorted(self._network.sinklist.items()):
            spec = OutputSpec(sink.id, "unknown", sink.datatype)
            self.outputs[spec.id] = self._OutputType(self, spec)

    def __eq__(self, other):
        """Compare two MacroNode instances with each other. This function ignores
        the parent and update status, but tests rest of the dict for equality.
        equality

        :param other: the other instances to compare to
        :type other: MacroNode
        :returns: True if equal, False otherwise
        """
        if not isinstance(other, MacroNode):
            return NotImplemented

        if self.id != other.id:
            return False

        if self.nodegroup != other.nodegroup:
            return False

        if self.inputs != other.inputs:
            return False

        if self.outputs != other.outputs:
            return False

        if self.network != other.network:
            return False

        return True

    def __getstate__(self):
        """
        Retrieve the state of the MacroNode

        :return: the state of the object
        :rtype dict:
        """
        state = super(MacroNode, self).__getstate__()
        state['inputs'] = []
        state['outputs'] = []
        state['network'] = self.network.__getstate__()
        return state

    def __setstate__(self, state):
        self._network = api.create_network_copy(state.pop('network')).parent
        super(MacroNode, self).__setstate__(state)
        self._create_inputs()
        self._create_outputs()
        self._output_info = {sink_id: None for sink_id in self._network.sinklist.keys()}

    def _update_input_mappings(self):
        """
        Update the mapping of the dimensions of source node to those
        of inputs. This is needed because the internal network has
        other dimensions than the macro node externally.
        """
        self._source_dimensions = {}
        self._source_sizes = {}
        for input_ in self.inputs.values():
            source_node = self.network.sourcelist[input_.id]

            if source_node.dimnames[0] not in self._source_dimensions:
                self._source_dimensions[source_node.dimnames[0]] = input_.dimnames
            if source_node.dimnames[0] not in self._source_sizes:
                self._source_sizes[source_node.dimnames[0]] = input_.size

    def get_output_info(self, output):
        """
        This functions maps the output dimensions based on the input dimensions
        of the macro. This is cached for speed as this can become rather costly
        otherwise

        :param output: output to get info for
        :return: tuple of Dimensions
        """
        if self._output_info[output.id] is None:

            # Translate back result index and id
            sink = self.network.sinklist[output.id]
            new_dimname = []
            new_size = []

            for dimname, size in zip(sink.dimnames, sink.outputsize):
                # If they were translated, replace them back
                if dimname in self._source_dimensions:
                    dimension_part = self._source_dimensions[dimname]
                    size_part = self._source_sizes[dimname]

                    new_dimname.extend(dimension_part)
                    new_size.extend(size_part)
                else:
                    new_dimname.append(dimname)
                    new_size.append(size)

            self._output_info[output.id] = tuple(Dimension(name, size) for name, size in zip(new_dimname, new_size))

        return self._output_info[output.id]

    def draw_link_target(self, context, port_name, input=True):
        if input:
            port_spec = 'i_{}'.format(port_name)
            if context['expand_macro'] is True or context['expand_macro'] >= len(context['network_stack']):
                return '{}_inputs:{}'.format(self.draw_id(context), port_spec)
        else:
            port_spec = 'o_{}'.format(port_name)
            if context['expand_macro'] is True or context['expand_macro'] >= len(context['network_stack']):
                return '{}_outputs:{}'.format(self.draw_id(context), port_spec)

        return '{}:{}'.format(self.draw_id(context), port_spec)

    def draw(self, context, graph, color=None):
        if color is None:
            color = context['colors']['macro_node']

        if context['expand_macro'] is True or context['expand_macro'] >= len(context['network_stack']):
            node_id = self.draw_id(context)

            context['network_stack'].append(self)
            old_dimensions = context['dimensions']
            context['dimensions'] = defaultdict(old_dimensions.default_factory)

            # Draw network
            with graph.subgraph(name='cluster_' + '_'.join(x.id for x in context['network_stack'])) as sub_graph:
                sub_graph.attr(label='{} ({} v{})'.format(self.id, self.network.id, self.network.version)),

                dimensions_in = ''
                dimensions_out = ''

                # Create connection node for inputs
                inputs = []
                outputs = []

                for input_ in self.inputs.values():
                    source_node = self.network.nodelist[input_.id]
                    if context['draw_dimensions']:
                        dimensions_in = '[' + 'x'.join(old_dimensions[x] for x in input_.dimnames) + ']'
                        dimensions_out = '[' + 'x'.join(context['dimensions'][x] for x in source_node.dimnames) + ']'

                    inputs.append('<i_{id}>{dim} {id}'.format(id=input_.id, dim=dimensions_in))
                    outputs.append('<dummy_{id}>{id} {dim}'.format(id=input_.id, dim=dimensions_out))
                inputs = '|'.join(inputs)
                outputs = '|'.join(outputs)

                sub_graph.node(
                    node_id + '_inputs',
                    label="<id>{id} inputs|{{{{{inputs}}}|{{{outputs}}}}}".format(id=self.id, inputs=inputs, outputs=outputs),
                    fillcolor=context['colors']['macro_link'],
                    style="filled"
                )

                # Create connection node for outputs
                inputs = []
                outputs = []

                for output in self.outputs.values():
                    sink_node = self.network.nodelist[output.id]
                    if context['draw_dimensions']:
                        dimensions_in = '[' + 'x'.join(context['dimensions'][x] for x in sink_node.dimnames) + ']'
                        dimensions_out = '[' + 'x'.join(old_dimensions[x] for x in output.dimnames) + ']'

                    inputs.append('<dummy_{id}>{dim} {id}'.format(id=output.id, dim=dimensions_in))
                    outputs.append('<o_{id}>{id} {dim}'.format(id=output.id, dim=dimensions_out))
                inputs = '|'.join(inputs)
                outputs = '|'.join(outputs)

                sub_graph.node(
                    node_id + '_outputs',
                    label="<id>{id} outputs|{{{{{inputs}}}|{{{outputs}}}}}".format(id=self.id, inputs=inputs, outputs=outputs),
                    fillcolor=context['colors']['macro_link'],
                    style="filled"
                )

                # Draw the macro network inside the sub_graph
                self.network.draw(context=context, graph=sub_graph)

                # Generate extra links to connect parent network to internal macro network
                for input_ in self.inputs.values():
                    source = '{}_inputs:dummy_{}'.format(node_id, input_.id)
                    target = self.network.nodelist[input_.id].draw_id(context)
                    graph.edge(source, target)

                for output in self.outputs.values():
                    source = self.network.nodelist[output.id].draw_id(context)
                    target = '{}_outputs:dummy_{}'.format(node_id, output.id)
                    graph.edge(source, target)

            # Reset context
            context['network_stack'].pop()
            context['dimensions'] = old_dimensions
        else:
            super(MacroNode, self).draw(context=context,
                                        graph=graph,
                                        color=color)

    def _update(self, key, forward=True, backward=False):
        """
        Update the Node information and validity of the Node and propagate
         the update downstream. Updates inputs, input groups, outputsize and outputs.

        A Node is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.planning.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized

        """
        super(MacroNode, self)._update(key, forward=False, backward=backward)
        self._update_input_mappings()
        self._output_info = {sink_id: None for sink_id in self._network.sinklist.keys()}

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, False)


class FlowNode(Node):
    """
    A Flow Node is a special subclass of Nodes in which the amount of samples
    can vary per Output. This allows non-default data flows.
    """
    _OutputType = Output

    def __init__(self, tool, id_=None, parent=None, resource_limits=None, nodegroup=None):
        """
        Instantiate a flow node.

        :param tool: The tool to base the node on
        :type tool: :py:class:`Tool <fastr.core.tool.Tool>`
        :param str id_: the id of the node
        :param parent: the parent network of the node
        :type parent: :py:class:`Network <fastr.planning.network.Network>`
        :return: the newly created FlowNode
        """
        super(FlowNode, self).__init__(tool, id_, parent=parent, resource_limits=resource_limits, nodegroup=nodegroup)

        self._input_groups = OrderedDict()

        # Update Inputs and self (which calls Outputs)
        self.update()

    @property
    def blocking(self):
        """
        A FlowNode is (for the moment) always considered blocking.

        :return: True
        """
        return True

    @property
    def outputsize(self):
        """
        Size of the outputs in this Node
        """
        # Get sizes of all input groups
        output_size = []
        for input_group in self.input_groups.values():
            if input_group.size is not None:
                output_size.extend(input_group.size)
            else:
                return None

        output_size.append(sympy.symbols('N_{}'.format(self.id)))
        return tuple(output_size)

    @property
    def dimensions(self):
        """
        Names of the dimensions in the Node output. These will be reflected
        in the SampleIdList of this Node.
        """
        if self.nodegroup is not None:
            extra_dim = self.nodegroup
        else:
            extra_dim = self.id

        extra_dim = Dimension(extra_dim, sympy.symbols('N_{}'.format(self.id)))

        return super(FlowNode, self).dimensions + (extra_dim,)


class AdvancedFlowNode(FlowNode):
    _OutputType = AdvancedFlowOutput


class SourceNode(FlowNode):
    """
    Class providing a connection to data resources. This can be any kind of
    file, stream, database, etc from which data can be received.
    """

    __dataschemafile__ = 'SourceNode.schema.json'
    _OutputType = SourceOutput

    def __init__(self, datatype, id_=None, parent=None, resource_limits=None, nodegroup=None):
        """
        Instantiation of the SourceNode.

        :param datatype: The (id of) the datatype of the output.
        :param id_: The url pattern.

        This class should never be instantiated directly (unless you know what
        you are doing). Instead create a source using the network class like
        shown in the usage example below.

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.create_network()
          >>> source = network.create_source(datatype=types['ITKImageFile'], id_='sourceN')
        """
        tool = resources.tools['fastr/Source:1.0', '1.0']

        super(SourceNode, self).__init__(tool, id_, parent=parent, resource_limits=resource_limits, nodegroup=nodegroup)

        self._input_groups = []

        # Set the DataType
        if datatype in types:
            if isinstance(datatype, str):
                datatype = types[datatype]
        else:
            message = 'Unknown DataType for SourceNode {} (found {}, which is not found in the types)!'.format(self.fullid, datatype)
            log.critical(message)
            raise exceptions.FastrValueError(message)

        self.datatype = datatype
        self._input_data = None

    def __getstate__(self):
        """
        Retrieve the state of the SourceNode

        :return: the state of the object
        :rtype dict:
        """
        state = super(SourceNode, self).__getstate__()

        return state

    def __setstate__(self, state):
        """
        Set the state of the SourceNode by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(SourceNode, self).__setstate__(state)

        self._input_data = None

    @property
    def datatype(self):
        """
        The datatype of the data this source supplies.
        """
        return self.outputs['output'].datatype

    @datatype.setter
    def datatype(self, value):
        """
        The datatype of the data this source supplies. (setter)
        """
        self.outputs['output'].datatype = value

    @property
    def sourcegroup(self):
        log.warning('[DEPRECATED] The sourcegroup property of the'
                          ' SourceNode is deprecated and replaced by the'
                          ' nodegroup property of the Node. Please use that'
                          ' property instead, it will have the same'
                          ' functionality')
        return self.nodegroup

    @property
    def nodegroup(self):
        return self._nodegroup

    @nodegroup.setter
    def nodegroup(self, value):
        self._nodegroup = value
        self._dimensions = None

    @property
    def dimensions(self):
        """
        The dimensions in the SourceNode output. These will be reflected
        in the SampleIdLists.
        """
        if self._dimensions is None:
            if self.nodegroup is not None:
                name = self.nodegroup
            else:
                name = self.id

            self._dimensions = Dimension(name, sympy.Symbol('N_{}'.format(name))),

        return self._dimensions

    @property
    def output(self):
        """
        Shorthand for ``self.outputs['output']``
        """
        return self.outputs['output']

    @property
    def valid(self):
        """
        This does nothing. It only overloads the valid method of Node().
        The original is intended to check if the inputs are connected to
        some output. Since this class does not implement inputs, it is skipped.
        """
        return True

    def draw(self, context, graph, color=None):
        if color is None:
            color = context['colors']['source_node']

        super(SourceNode, self).draw(context=context,
                                     graph=graph,
                                     color=color)

    def set_data(self, data, ids=None):
        """
        Set the data of this source node.

        :param data: the data to use
        :type data: dict, OrderedDict or list of urls
        :param ids: if data is a list, a list of accompanying ids
        """
        self._input_data = OrderedDict()

        # Check if data has key or generate keys
        log.debug('Storing {} (ids {}) in {}'.format(data, ids, self.fullid))
        if isinstance(data, OrderedDict):
            ids = [SampleId(x) for x in data.keys()]
            data = list(data.values())
        elif isinstance(data, dict):
            # Have data sorted on ids
            ids, data = list(zip(*sorted(data.items())))
            ids = [SampleId(x) for x in ids]
        elif isinstance(data, list):
            if ids is None:
                ids = [SampleId('id_{}'.format(k)) for k in range(len(data))]
            elif not isinstance(ids, list):
                raise exceptions.FastrTypeError('Invalid type! The ids argument should be a list that matches the data samples!')
        elif isinstance(data, tuple):
            # A single sample with cardinality
            ids = [SampleId('id_0')]
            data = [data]
        else:
            if isinstance(data, set):
                log.warning('Source data for {} is given as a set,'.format(self.fullid) +
                                  ' this is most probably a mistake and a list or dict should'
                                  ' be used instead')
            ids = [SampleId('id_0')]
            data = [data]

        for key, value in zip(ids, data):
            if isinstance(value, tuple):
                self._input_data[key] = tuple(x if self.datatype.isinstance(x) else types['String'](str(x)) for x in value)
            else:
                self._input_data[key] = (value if self.datatype.isinstance(value) else types['String'](str(value))),
            log.debug('Result {}: {} (Type {})'.format(key, self._input_data[key], type(self._input_data[key]).__name__))

    def _update(self, key, forward=True, backward=False):
        """
        Update the Node information and validity of the Node and propagate
         the update downstream. Updates inputs, input_groups, outputsize and outputs.

        A Node is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.planning.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized

        """
        # Make sure the Inputs and input groups are up to date
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        for input_ in self.inputs.values():
            input_.update(key)

        self.update_input_groups()

        # Update own status
        valid = True
        messages = []

        if len(self.listeners) == 0:
            valid = False
            messages.append('[{}] No nodes are linked to any output of this SourceNode'.format(self.id))

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('Input {} is not valid: {}'.format(id_, message))

        for input_group in self.input_groups.values():
            if input_group.empty:
                valid = False
                messages.append('InputGroup {} is empty'.format(input_group.id))

        for id_, output in self.outputs.items():
            if output.resulting_datatype is not None and not issubclass(output.resulting_datatype, DataType):
                    valid = False
                    messages.append(
                        'Output {} cannot determine the Output DataType (got {}), '
                        'please specify a valid DataType or add casts to the Links'.format(
                            id_, output.resulting_datatype))

        self._status['valid'] = valid
        self._status['messages'] = messages

        # Update all outputs
        for output in self.outputs.values():
            output.update(key)

        # Update all downstream listeners
        if forward:
            for listener in self.listeners:
                listener.update(key, forward, backward)


class SinkNode(Node):
    """
    Class which handles where the output goes. This can be any kind of file, e.g.
    image files, textfiles, config files, etc.
    """

    __dataschemafile__ = 'SinkNode.schema.json'

    def __init__(self, datatype, id_=None, parent=None, resource_limits=None, nodegroup=None):
        """ Instantiation of the SourceNode.

        :param datatype: The datatype of the output.
        :param id_: the id of the node to create
        :return: newly created sink node

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.create_network()
          >>> sink = network.create_sink(datatype=types['ITKImageFile'], id_='SinkN')

        """
        Node.__init__(self, resources.tools['fastr/Sink:1.0', '1.0'], id_, parent=parent, resource_limits=resource_limits, nodegroup=nodegroup)
        # Set the DataType
        if datatype in types:
            if isinstance(datatype, str):
                datatype = types[datatype]
        else:
            message = 'Invalid DataType for SinkNode {} (found {})!'.format(self.fullid, datatype)
            log.critical(message)
            raise exceptions.FastrValueError(message)

        self.datatype = datatype
        self.url = None

    def __getstate__(self):
        state = super(SinkNode, self).__getstate__()
        state['url'] = self.url
        return state

    def __setstate__(self, state):
        super(SinkNode, self).__setstate__(state)
        self.url = state['url']

    @property
    def datatype(self):
        """
        The datatype of the data this sink can store.
        """
        return self.inputs['input'].datatype

    @datatype.setter
    def datatype(self, value):
        """
        The datatype of the data this sink can store (setter).
        """
        self.inputs['input'].datatype = value

    @property
    def input(self):
        """
        The default input of the sink Node
        """
        return self.inputs['input']

    @input.setter
    def input(self, value):
        """
        The default input of the sink Node (setter)
        """
        self.inputs['input'] = value

    def draw(self, context, graph, color=None):
        if color is None:
            color = context['colors']['sink_node']

        super(SinkNode, self).draw(context=context,
                                     graph=graph,
                                     color=color)

    def _update(self, key, forward=True, backward=False):
        """
        Update the Node information and validity of the Node and propagate
         the update downstream. Updates inputs, input groups, outputsize and outputs.

        A Node is valid if:

        * All Inputs are valid (see :py:meth:`Input.update <fastr.planning.inputoutput.Input.update>`)
        * All InputGroups are non-zero sized

        """
        # Make sure the Inputs and input groups are up to date
        # log.debug('Update {} passing {} {}'.format(key, type(self).__name__, self.id))

        if backward:
            for sourced_node in self.get_sourced_nodes():
                sourced_node.update(key, False, backward)

        for input_ in self.inputs.values():
            input_.update(key, forward, backward)

        # Reset dimensions cache
        self._dimensions = None

        self.update_input_groups()
        self.input_group_combiner.update()

        # Update own status
        valid = True
        messages = []

        if len(self.get_sourced_nodes()) == 0:
            valid = False
            messages.append('[{}] No nodes are linked to any input of this Node'.format(self.id))

        for id_, input_ in self.inputs.items():
            if not input_.valid:
                valid = False
                for message in input_.messages:
                    messages.append('[{}] Input {} is not valid: {}'.format(self.id, input_.id, message))

        for input_group in self.input_groups.values():
            if input_group.empty:
                valid = False
                messages.append('[{}] InputGroup {} is empty'.format(self.id, input_group.id))

        self._status['valid'] = valid
        self._status['messages'] = messages


class ConstantNode(SourceNode):
    """
    Class encapsulating one output for which a value can be set. For example
    used to set a scalar value to the input of a node.
    """

    __dataschemafile__ = 'ConstantNode.schema.json'

    def __init__(self, datatype, data, id_=None, parent=None, resource_limits=None, nodegroup=None):
        """
        Instantiation of the ConstantNode.

        :param datatype: The datatype of the output.
        :param data: the prefilled data to use.
        :param id_: The url pattern.

        This class should never be instantiated directly (unless you know what
        you are doing). Instead create a constant using the network class like
        shown in the usage example below.

        usage example:

        .. code-block:: python

          >>> import fastr
          >>> network = fastr.create_network()
          >>> source = network.create_source(datatype=types['ITKImageFile'], id_='sourceN')

        or alternatively create a constant node by assigning data to an item in an InputDict:

        .. code-block:: python

          >>> node_a.inputs['in'] = ['some', 'data']

        which automatically creates and links a ConstantNode to the specified Input
        """
        super(ConstantNode, self).__init__(datatype, id_, parent=parent, resource_limits=resource_limits, nodegroup=nodegroup)
        self.set_data(data)
        self._data = self._input_data

    def __getstate__(self):
        """
        Retrieve the state of the ConstantNode

        :return: the state of the object
        :rtype dict:
        """
        state = super(ConstantNode, self).__getstate__()

        state['data'] = [[k.__getnewargs__(), [x.serialize() for x in v]] for k, v in self._data.items()]

        return state

    def __setstate__(self, state):
        """
        Set the state of the ConstantNode by the given state.

        :param dict state: The state to populate the object with
        :return: None
        """
        super(ConstantNode, self).__setstate__(state)

        self._data = OrderedDict((SampleId(*key), tuple(DataType.deserialize(x) for x in value)) for key, value in state['data'])
        self.set_data()  # Make sure that the output size etc gets set

    def set_data(self, data=None, ids=None):
        """
        Set the data of this constant node in the correct way. This is mainly
        for compatibility with the parent class SourceNode

        :param data: the data to use
        :type data: dict or list of urls
        :param ids: if data is a list, a list of accompanying ids
        """
        # We have to arguments to match the superclas
        # pylint: disable=unused-argument
        if data is None and self.data is not None:
            self._input_data = self.data
        else:
            super(ConstantNode, self).set_data(data, ids)

    @property
    def data(self):
        """
        The data stored in this constant node
        """
        return self._data

    @property
    def print_value(self):
        result = []
        for value in self.data.values():
            value = [str(x) for x in value]
            value = tuple((x[:12] + '...' + x[-18:]) if len(x) > 33 else x for x in value)
            result.append(value)

        return pprint.pformat(result, width=36).replace('\n', r'\n')

    def draw(self, context, graph, color=None):
        if color is None:
            color = context['colors']['constant_node']

        graph.node(self.draw_id(context=context),
                   label="<id>{id}|<o_output>{data}".format(id=self.id, data=self.print_value),
                   fillcolor=color,
                   style="filled")



# Copyright 2011-2014 Biomedical Imaging Group Rotterdam, Departments of
# Medical Informatics and Radiology, Erasmus MC, Rotterdam, The Netherlands
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The datatypes module holds all DataTypes generated by fastr and all the base
classes for these datatypes.
"""

# Empty module to be populated by the fastr.plugins.managers.datatypemanager.DataTypeManager
from abc import abstractmethod, abstractproperty
import os
import sys
import traceback
from typing import Generic, Iterable, Optional, Sequence, Tuple, Type, Union
import urllib.parse
import xml.etree.ElementTree as ElementTree

from .. import exceptions
from ..abc.baseplugin import BasePlugin, PluginState
from ..abc.basepluginmanager import BasePluginManager
from ..abc.serializable import Serializable, load
from ..core.version import Version
from ..core import vfs_plugin
from ..data import url
from ..helpers import config, log
from ..helpers.checksum import md5_checksum, hashsum
from ..helpers.classproperty import classproperty


# Placeholder until the actually datatype manager is running
types: 'DataTypeManager'


# Basic functions for pickle how to find DataTypes that are pickled
# these are referenced from __reduce_ex__
def _get_type(id_: str) -> 'BaseDataType':
    obj = types[id_]()
    return obj


def _get_enum(id_: str, options: Iterable[str]) -> 'EnumType':
    enum_type = types.create_enumtype(type_id=id_, options=options)
    return enum_type()


class BaseDataType(BasePlugin):
    """
    The base class for all datatypes in the fastr type system.
    """
    filename: str = __file__

    #: Version of the DataType definition
    version: Version = Version('1.0')

    #: Description of the DataType
    description: str = ''

    #: Extension related to the Type
    extension = None

    # DataTypes do not need to be loaded, so they are always Loaded
    _status = (PluginState.loaded, 'DataTypes are always loaded', '')

    @abstractmethod
    def __init__(self, value=None, format_: str=None):
        """
        The BaseDataType constructor.

        :param value: value to assign to the new BaseDataType object
        :param format_: the format used for the ValueType
        :return: new BaseDataType object
        :raises FastrNotImplementedError: if *id*, *name*, *version* or *description* is None
        """
        super(BaseDataType, self).__init__()

        self._value = None
        self.value = value
        self.format = format_

    def __repr__(self) -> str:
        """
        Returns string representation of the BaseDataType

        :return: string represenation
        :rtype: str
        """
        return "<{}: {}>".format(self.id, repr(self.value))

    def __str__(self) -> str:
        """
        Returns the string version of the BaseDataType

        :return: string version
        :rtype: str
        """
        return str(self.value)

    def __eq__(self, other: 'BaseDataType') -> bool:
        """
        Test the equality of two DataType objects

        :parameter DataType other: the object to compare against
        :return: flag indicating equality
        :rtype: bool
        """
        if not isinstance(self, type(other)):
            return NotImplemented

        return self.value == other.value

    def __ne__(self, other: 'BaseDataType') -> bool:
        """
        Test if two objects are not equal. This is by default done by negating
        the __eq__ operator

        :parameter DataType other: the object to compare against
        :return: flag indicating equality
        :rtype: bool
        """
        if not isinstance(self, type(other)):
            return NotImplemented

        return not self.__eq__(other)

    def __reduce_ex__(self, *args, **kwargs):
        return _get_type, (self.id,), self.__getstate__()

    def __getstate__(self):
        return type(self).id, self._value, self.format

    def __setstate__(self, state):
        if self.id != state[0]:
            raise exceptions.FastrValueError('Invalid state for {}, state is for type {}'.format(self.id, state[0]))

        self._value = state[1]
        self.format = state[2]

    @classproperty
    def id(cls) -> str:
        """
        Internal id used for the DataType
        """
        return cls.__name__

    @classproperty
    def dot_extension(cls):
        """
        Extension(s) with a prefixed dot
        """
        if not cls.extension:
            return

        if isinstance(cls.extension, str):
            return '.{}'.format(cls.extension)
        else:
            # Must be tuple/list
            return tuple('.{}'.format(x) for x in cls.extension)

    @classproperty
    def fullid(cls):
        """
        The full fastr id of the DataType
        """
        return '{}/{}'.format(cls.parent.fullid, cls.id)

    @classproperty
    def name(cls):
        """
        Display friendly name of the DataType
        """
        return cls.id

    @property
    def value(self):
        """
        The value of object instantiation of this DataType.
        """
        return self._value

    @property
    def raw_value(self):
        """
        The raw value of object instantiation of this DataType. For datatypes
        that override value (like Deferred) this is the way to access the
        _value field.
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Setter function for value property
        """
        if isinstance(value, BaseDataType):
            if self.isinstance(value):
                self._value = value.value
                self.format = value.format
            else:
                raise exceptions.FastrTypeError('Cannot create a {} based on a {} (non-matching datatypes)'.format(
                    type(self).id,
                    type(value).id
                ))
        else:
            self._value = value

    @classproperty
    def parent(self):
        """
        The parent container of the DataType
        """
        return types

    @property
    def parsed_value(self):
        """
        The parsed value of object instantiation of this DataType.
        """
        return self._value

    @classmethod
    def test(cls):
        """
        Define the test for the BasePluginManager. Make sure we are not one of the
        base classes
        """
        if cls in [BaseDataType, DataType, TypeGroup, EnumType, ValueType, URLType]:
            raise exceptions.FastrTypeError('Cannot use a basic datatype {} as a plugin!'.format(cls))

    @property
    def valid(self):
        """
        A boolean flag that indicates weather or not the value assigned to this
        DataType is valid. This property is generally overwritten by
        implementation of specific DataTypes.
        """
        # Avoid errors in the validation testing (this is external code)
        # so we use a broad except on purpose
        # pylint: disable=broad-except
        try:
            return self._validate()
        except Exception:
            exc_type, _, _ = sys.exc_info()
            exc_info = traceback.format_exc()
            log.warning('Could not validate {}: encountered exception ({}) during execution:\n{}'.format(repr(self), exc_type.__name__, exc_info))
            raise

    def _validate(self):
        """
        The actual validation function to be overwritten by subclasses.
        """
        # This function doesn't use self, but is intended to be potentially
        # overwritten by the subclasses.
        # pyline: disbale=no-self-use
        return True

    def checksum(self):
        """
        Generate a checksum for the value of this DataType

        :return: the checksum of the value
        :rtype: str
        """
        return hashsum(self.value)

    @classmethod
    def isinstance(cls, value):
        """
        Indicate whether value is an instance for this DataType.

        :return: the flag indicating the value is of this DataType
        :rtype: bool
        """
        return isinstance(value, cls)


class DataType(BaseDataType, Serializable):
    """
    This class is the base class for all DataTypes that can hold a value.
    """
    @abstractmethod
    def __init__(self, value=None, format_=None):
        """
        The DataType constructor.

        :param value: value to assign to the new DataType object
        :param format: the format used for the ValueType
        :return: new DataType object
        """
        super(DataType, self).__init__(value, format_)

    def serialize(self) -> dict:
        """
        Method that returns a dict structure with the datatype
        the object.

        :returns: serialized representation of object
        """
        return {
            'id': self.id,
            'format': self.format,
            'value': self._value,
        }

    @classmethod
    def deserialize(cls, doc: dict, _=None) -> 'DataType':
        """
        Classmethod that returns an object constructed based on the
        str/dict (or OrderedDict) representing the object

        :param doc: the state of the object to create
        :return: newly created object (of datatype indicated by the doc)
        """
        datatype = types[doc['id']]
        obj = datatype(value=doc['value'], format_=doc.get('format'))
        return obj

    def action(self, name):
        """
        This function can be overwritten by subclasses to implement certain
        action that should be performed. For example, the *Directory* DataType
        has an action *ensure*. This method makes sure the Directory exists.
        A Tool can indicate an action that should be called for an Output which
        will be called before execution.

        :param str name: name of the action to execute
        :return: None
        """
        if name is not None:
            log.warning("unknown action '{}' for DataType {}".format(name, self.id))


class Missing(DataType):
    """
    Singleton DataType to annotate missing data
    """
    _instance = None
    _initialized = False
    value = 'MISSING'

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls._instance, cls):
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self, _=None, __=None):
        if not self._initialized:
            super().__init__('MISSING', None)


class TypeGroup(BaseDataType):
    """
    The TypeGroup is a special DataType that does not hold a value of its own
    but is used to group a number of DataTypes. For example ITK has a list of
    supported file formats that all tools build on ITK support. A group can
    be used to conveniently specify this in multiple Tools that use the same
    set DataTypes.
    """

    _member_types = None
    _preference = None

    def __new__(cls, value=None, format_=None):
        """
        Instantiate a TypeGroup. This will for match the value to the best
        matching type and instantiate that. Not that the returned object
        will not be of type TypeGroup but one of the TypeGroup members.
        """
        # Avoid casting values that are already in a member type
        if any(isinstance(value, x) for x in cls.members):
            return value

        matching_type = types.guess_type(value, options=cls)

        # Only continue if we have value urls
        if isinstance(value, str) and value.startswith('val://'):
            if matching_type is None:
                # Just try a simple match rather than guessing
                matching_type = types.match_types(cls, type(value))

            if matching_type is None:
                # Just try a simple match rather than guessing
                matching_type = types.match_types(cls)

        if matching_type is None:
            raise exceptions.FastrValueError('Cannot matching value {} [{}] to any of {}'.format(value,
                                                                                                 type(value).__name__,
                                                                                                 cls.members))

        return matching_type(value, format_)

    def __init__(self, value=None):
        """
        Dummy constructor. TypeGroups are not instantiable and cannot hold a
        value of its own.

        :raises FastrDataTypeNotInstantiableError: if called
        """
        # All type groups are per definition not instantiable
        # pylint: disable=super-init-not-called
        raise exceptions.FastrDataTypeNotInstantiableError('TypeGroups are not instantiable')

    @property
    @abstractmethod
    def _members(self):
        """
        The id of the members, this should be set in the subclass. It should
        be a frozen set of str.
        """

    @classmethod
    def _member_list(cls):
        members = tuple(types[x] for x in cls._members if x in types)
        members = tuple(x.members if issubclass(x, TypeGroup) else (x,) for x in members)
        members = tuple(x for y in members for x in y)
        return members

    @classproperty
    def members(cls):
        """
        The members of the TypeGroup
        """
        if cls._member_types is None:
            cls._member_types = frozenset(cls._member_list())

        return cls._member_types

    @classproperty
    def preference(cls):
        if cls._preference is None:
            if hasattr(cls, "preferred_types"):
                cls._preference = tuple(types[x] for x in cls.preferred_types)
            else:
                preference = [x.preference if issubclass(x, TypeGroup) else (x,) for x in cls._member_list()]
                cls._preference = tuple(x for y in preference for x in y)

        return cls._preference

    @classmethod
    def isinstance(cls, value):
        return any(x.isinstance(value) for x in cls.members)


class AnyType(TypeGroup):
    """
    Special Datatype in fastr that is a TypeGroup with all known DataTypes as
    its members.
    """
    @classproperty
    def _members(cls):
        """
        A "class-poperty" that gives a list of the ids of all currently loaded DataTypes
        """
        # During object construction this is executed and types does not yet exist
        try:
            types_list = types
        except NameError:
            types_list = []

        return frozenset(x.id for x in types_list.values() if issubclass(x, DataType))

    @classproperty
    def description(cls):
        """
        The description of the AnyType, including the list of member types.
        """
        disp_members = ['  - {}'.format(member) for member in cls.members]
        return """
TypeGroup {id}
{name} ({id}) is a group of consisting of all DataTypes known by fastr, currently:
{members}
""".strip().format(id='AnyType', name='AnyType', members='\n'.join(disp_members))


class AnyFile(TypeGroup):
    """
    Special Datatype in fastr that is a TypeGroup with all known DataTypes as
    its members.
    """

    @classproperty
    def _members(cls):
        """
        A "class-poperty" that gives a list of the ids of all currently loaded DataTypes
        """
        # During object construction this is executed and types does not yet exist
        try:
            types_list = types
        except NameError:
            types_list = []

        return frozenset(x.id for x in types_list.values() if issubclass(x, URLType))

    @classproperty
    def description(cls):
        """
        The description of the AnyType, including the list of member types.
        """
        disp_members = ['  - {}'.format(member) for member in cls.members]
        return """
TypeGroup {id}
{name} ({id}) is a group of consisting of all URLTypes known by fastr, currently:
{members}
""".strip().format(id='AnyFile', name='AnyFile', members='\n'.join(disp_members))


class EnumType(DataType):
    """
    The EnumType is the base for DataTypes that can have a value which is an
    option from a predefined set of possibilities (similar to an enum type
    in many programming languages).
    """
    #: Enums always have version 1.0
    version = Version('1.0')
    _options = frozenset()

    def __init__(self, value=None, format_=None):
        """
        The EnumType constructor.

        :param value: value to assign to the new EnumType object
        :param format: the format used for the ValueType
        :return: new EnumType object
        :raises FastrDataTypeNotInstantiableError: if not subclassed
        """
        super(EnumType, self).__init__(value, format_)

        if self.__class__ is EnumType:
            raise exceptions.FastrDataTypeNotInstantiableError('EnumType is not instantiable')

    def __reduce_ex__(self, *args, **kwargs):
        return _get_enum, (self.id, self.options), self.__getstate__()

    @classproperty
    def description(cls):
        """
        The description of the AnyType, including the list of member types.
        """
        disp_options = ['  - "{}"'.format(option) for option in cls.options]
        return """
{name} ({id}) is a enumerate type with options:
{options}

{name} can take the value of any of the option, but any other value is considered invalid.
""".strip().format(id=cls.id, name=cls.name, options='\n'.join(disp_options))

    @classproperty
    def options(cls):
        """
        A frozenset holding the options that the value of the EnumType object can have.

        :return: the options the value can hold
        :rtype: frozenset
        """
        return cls._options

    def _validate(self):
        return self._value in self._options


class ValueType(DataType):
    """
    The ValueType is the base for DataTypes that hold simple values (not an
    EnumType and not a file/URL). The values is generally represented by a
    string.
    """

    def __init__(self, value=None, format_=None):
        """
        The ValueType constructor

        :param value: value to assign to the new ValueType
        :param format: the format used for the ValueType
        :return: new ValueType object
        """
        super(ValueType, self).__init__(value, format_)


class URLType(DataType):
    """
    The URLType is the base for DataTypes that point to a resource somewhere
    else (typically a filesystem). The true value is actually the resource
    referenced by the value in this object.
    """

    def __init__(self, value=None, format_=None):
        """
        The URLType constructor

        :param value: value to assign to the new URLType
        :param format: the format used for the ValueType
        :return: new URLType object
        """
        super(URLType, self).__init__(value, format_)

    def __eq__(self, other):
        """
        Test the equality of two DataType objects

        :parameter URLType other: the object to compare against
        :return: flag indicating equality
        :rtype: bool
        """
        if not isinstance(self, type(other)):
            return NotImplemented

        return self.checksum() == other.checksum()

    def checksum(self):
        """
        Return the checksum of this URL type

        :return: checksum string
        :rtype: str
        """
        contents = self.content(self.parsed_value)
        return md5_checksum(contents)

    @property
    def parsed_value(self):
        """
        The parsed value of object instantiation of this DataType.
        """
        if url.isurl(self.value):
            parsed_url = urllib.parse.urlparse(self.value)

            if parsed_url.scheme == 'vfs':
                return vfs_plugin.url_to_path(self.value)
            else:
                raise exceptions.FastrValueError('Cannot get parsed value for non-vfs url: {} (scheme {})'.format(self.value, parsed_url.scheme))
        else:
            return self.value

    @classmethod
    def content(cls, inval, outval=None):
        """
        Give the contents of a URLType, this is generally useful for filetypes
        that consists of multiple files (e.g. AnalyzeImageFile, DICOM). The
        value will indicate the main file, and the contents function can
        determine all files that form a single data value.

        :param inval: a value to figure out contents for this type
        :param outval: the place where the copy should point to
        :return: a list of all files part of the value (e.g. header and data file)
        :rtype: list
        """
        if outval is not None:
            return [(inval, outval)]
        else:
            return [inval]

    @property
    def valid(self):
        """
        A boolean flag that indicates weather or not the value assigned to this
        DataType is valid. This property is generally overwritten by
        implementation of specific DataTypes.
        """
        if not isinstance(self.value, str):
            return False

        return super(URLType, self).valid

    def _validate(self):
        """
        The actual validation function to be overwritten by subclasses.
        """
        if self.extension and not self.parsed_value.endswith(self.dot_extension):
            return False

        try:
            contents = self.content(self.parsed_value)

            for content in contents:
                if not os.path.exists(content):
                    return False
            return True
        except (TypeError, ValueError, IOError):
            return False


class Deferred(DataType):
    def __init__(self, value=None, format_=None):
        """
        The Deferred constructor.

        :param value: value to assign to the new DataType object
        :param format: This is ignore but here for compatibility
        :return: new Deferred object
        """
        self._value = value
        self._data = None
        self.format = format_

    def __repr__(self):
        """
        Returns string representation of the BaseDataType

        :return: string represenation
        :rtype: str
        """
        if self.target is None:
            return "<{}: {}>".format(self.id, repr(self._value))
        else:
            return "<{}: {}>".format(self.id, repr(self.target))

    def __getstate__(self):
        return ('Deferred', self._value)

    def __setstate__(self, state):
        if self.id != state[0]:
            raise exceptions.FastrValueError('Unvalid state for {}, state is for type {}'.format(self.id, state[0]))

        self._value = state[1]

    @classmethod
    def lookup(cls, value):
        """
        Look up the deferred target and return that object

        :param: value
        :return: The value the deferred points to
        :rtype: DataType
        :raises FastrKeyError: if the deferred is not available (yet)
        :raises FastrValueError: if the value is not a valid deferrred url
        """
        parsed_url = urllib.parse.urlparse(value)

        if parsed_url.scheme == 'val':
            datafile = os.path.join(config.mounts[parsed_url.netloc], os.path.normpath(parsed_url.path[1:]))
            query = urllib.parse.parse_qs(parsed_url.query)

            # Open Job file
            data = load(datafile)

            # Attempt to extract data
            try:
                outputname = query['outputname'][0]
                cardinality_nr = int(query['nr'][0])
                if 'sampleid' in query:
                    sample_id = query['sampleid'][0]
                    value = data.output_data[outputname][sample_id][cardinality_nr]
                else:
                    value = data.output_data[outputname][cardinality_nr]
            except (IndexError, KeyError) as exception:
                log.debug('Output data for query: {}'.format(data.output_data))
                message = 'Could not get value from {}, encountered {}: {}'.format(value,
                                                                                   type(exception).__name__,
                                                                                   exception.args[0])
                raise exceptions.FastrKeyError(message)
        else:
            raise exceptions.FastrValueError('Cannot lookup value {}, wrong url scheme'.format(value))

        if isinstance(value, Deferred):
            value = value.target

        return value

    @property
    def target(self):
        """
        Target object for this deferred.

        :raises FastrKeyError: if the deferred is not available (yet)
        :raises FastrValueError: if the value is not a valid deferrred url
        """
        if self._data is None:
            try:
                self._data = self.lookup(self._value)
            except (exceptions.FastrKeyError,
                    exceptions.FastrValueError,
                    exceptions.FastrFileNotFound):
                pass

        return self._data

    @property
    def value(self):
        """
        The value of object instantiation of this DataType.
        """
        try:
            target = self.target
        except exceptions.FastrKeyError:
            return None

        if target is None:
            return None

        return target.value

    @property
    def parsed_value(self):
        """
        The value of object instantiation of this DataType.
        """
        try:
            target = self.target
        except exceptions.FastrKeyError:
            return None

        if target is None:
            return None

        return target.parsed_value

    @property
    def provenance(self):
        return self._get_data()[1]

    @property
    def job(self):
        try:
            return self._get_data()[2]
        except exceptions.FastrValueError:
            return None

    def _validate(self):
        """
        The actual validation function to be overwritten by subclasses.
        """
        # This function doesn't use self, but is intended to be potentially
        # overwritten by the subclasses.
        # pyline: disbale=no-self-use
        target = self.target

        if target is None:
            return False

        return target.valid

    def checksum(self):
        """
        Generate a checksum for the value of this DataType

        :return: the checksum of the value
        :rtype: str
        """
        target = self.target

        if target is None:
            return hashsum('__FASTR_NOT_AVAILABLE_HASH__')

        return target.checksum()


def fastr_isinstance(obj, datatype):
    """
    Check if an object is of a specific datatype.

    :param obj: Object to inspect
    :param datatype: The datatype(s) to check
    :type datatype: tuple, BaseDataType
    :return: flag indicating object is of datatype
    :rtype: bool
    """
    if not isinstance(datatype, tuple):
        datatype = datatype,

    for dtype in datatype:
        if issubclass(dtype, TypeGroup):
            if type(obj) in dtype.members:
                return True
        elif isinstance(obj, datatype):
            return True

    return False


DataTypeClass = Type[BaseDataType]

class DataTypeManager(BasePluginManager[DataTypeClass]):
    """
    The DataTypeManager hold a mapping of all DataTypes in the fast system and
    can create new DataTypes from files/data structures.
    """

    def __init__(self):
        """
        The DataTypeManager constructor will create a new DataTypeManager and
        populate it with all DataTypes it can find in the paths set in
        ``config.types_path``.

        :return: the created DataTypeManager
        """
        self.types_map = {}
        super(DataTypeManager, self).__init__(config.types_path)

    @property
    def preferred_types(self):
        return [self.data[t] for t in config.preferred_types if t in self.data]

    @property
    def fullid(self):
        """
        The fullid of the datatype manager
        """
        return 'fastr://types'

    @property
    def plugin_class(self):
        """
        The PluginClass of the items of the BasePluginManager
        """
        return BaseDataType

    # Allow key to be a id string or DataType
    def __keytransform__(self, key):
        """
        Key transformation for this mapping. The key transformation allows
        indexing by both the DataType name as well as the DataType it self.

        :param key: The name of the requested datatype or the datatype itself
        :type key: fastr.datatypes.BaseDataType or str
        :return: The requested datatype
        """
        if self.isdatatype(key):
            if key.name in self.data and self.data[key.name] is key:
                return key.name
            else:
                raise exceptions.FastrDataTypeMismatchError('key DataType {} not {}'.format(key.name, type(self).__name__))
        else:
            return key

    def populate(self):
        """
        Populate Manager. After scanning for DataTypes, create the AnyType and set the preferred types
        """
        super(DataTypeManager, self).populate()

        # Add the any type
        self['AnyType'] = AnyType
        self['AnyFile'] = AnyFile
        self['Deferred'] = Deferred
        self['Missing'] = Missing

    @property
    def _instantiate(self):
        """
        Flag indicating that the plugin should NOT be instantiated prior to saving
        """
        return False

    def _print_key(self, key):
        if key.startswith('__') and key.endswith('__'):
            return None

        return key

    def has_type(self, name):
        """
        Check if the datatype with requested name exists

        :param str name: the name of the requested datatype
        :return: flag indicating if the datatype exists
        :rtype: bool
        """
        return name in self.data

    def poll_datatype(self, filename):
        """
        Poll an xml file to see if there is a definition of a datatype in it.

        :param str filename: path of the file to poll
        :return: tuple with (id, version, basetype) if a datatype is found or (None, None, None) if no datatype is found
        """
        if os.path.exists(filename):
            tree = ElementTree.parse(filename)
            root = tree.getroot()

            if root.tag not in ('type', 'typegroup'):
                message = 'Invalid root tag ({}) in file!'.format(root.tag)
                log.warning(message)
                return (None, None, None)

            id_ = root.get('id')
            version = Version(root.get('version'))

            return (id_, version, root.tag)
        else:
            message = '{} not a valid filename'.format(filename)
            log.warning(message)
            return (None, None, None)

    def get_type(self, name: str) -> DataTypeClass:
        """Read a type given a typename. This will scan all directories in
        types_path and attempt to load the newest version of the DataType.

        :param str name: Name of the datatype that should be imported in the system
        :return: the datatype with the requested name, or None if datatype is not found

        .. note:: If type is already in TypeManager it will not load anything and return the
                  already loaded version.
        """
        log.debug('Attemping to get datatype {}'.format(name))
        if name in self:
            return self[name]

        latest_version = Version('0.0')
        latest_filename = ''

        for path in config.types_path:
            filename = os.path.join(path, name + '.xml')
            if os.path.exists(filename):
                (pollname, version, _) = self.poll_datatype(filename)
                if pollname == name and version > latest_version:
                    latest_version = version
                    latest_filename = filename

        if latest_filename == '':
            message = 'Could not find type with name {}'.format(name)
            log.error(message)
            return None

        log.debug('Found {} (version {}) in {}'.format(name, latest_version, latest_filename))
        self._load_item(latest_filename)
        return self[name]

    def _store_item(self, name, value):
        """
        Store an item in the BaseManager, will ignore the item if the key is
        already present in the BaseManager.

        :param name: the key of the item to save
        :param value: the value of the item to save
        :return: None
        """
        super(DataTypeManager, self)._store_item(name, value)

        if value.id is not None:
            value.__module__ = 'fastr.datatypes'
            setattr(types, value.id, value)

    def create_enumtype(self, type_id: str, options: Iterable[str], name: str=None) -> Type[EnumType]:
        """
        Create a python class based on an XML file. This function return a
        completely functional python class based on the contents of a DataType
        XML file.

        Such a class will be of type EnumType.

        :param str type_id: the id of the new class
        :param iterable options: an iterable of options, each option should be str
        :return: the newly created subclass of EnumType
        :raises FastrTypeError: if the options is not an iterable of str
        """
        if type_id in self:
            if self[type_id].options != set(options):
                raise exceptions.FastrDataTypeMismatchError('Conflicting definition of Enum {}!'
                                                            ' (options {} vs {})'.format(type_id,
                                                                                         self[type_id].options,
                                                                                         options))

            log.debug('Returning existing DataType {}!'.format(type_id))
            return self[type_id]

        attributes = {}

        try:
            if isinstance(options, str):
                options = (options,)

            attributes['_options'] = frozenset(options)
        except TypeError:
            message = 'options must be a iterable containing the valid options for the Enum, found options {}'.format(options)
            log.error(message)
            raise exceptions.FastrTypeError(message)

        if not all(isinstance(x, str) for x in attributes['_options']):
            message = 'all options for an Enum must be of type str, found options {}'.format(options)
            log.error(message)
            raise exceptions.FastrTypeError(message)

        attributes['parent'] = self
        attributes['_sourcepath'] = None
        attributes['_hash'] = hashsum([type_id, name, options])
        attributes['__module__'] = 'fastr.datatypes'

        supertypes = (EnumType,)

        log.debug('Creating EnumType {} from script'.format(type_id))
        out: Type[EnumType] = type(type_id, supertypes, attributes)
        self[type_id] = out
        setattr(types, out.id, out)
        return out

    def guess_type(self,
                   value: str,
                   exists: Optional[bool] = True,
                   options: Optional[Union[DataTypeClass, Tuple[DataTypeClass]]] = None,
                   preferred: Optional[Sequence[DataTypeClass]] = None) -> Optional[DataTypeClass]:
        """
        Guess the DataType based on a value str.

        :param str value: the value to guess the type for
        :param options: The options that are allowed to be guessed from
        :type options: TypeGroup, DataType or tuple of DataTypes
        :param bool exists: Indicate the value exists (if file) and can be
                             checked for validity, if false skip validity check
        :param iterable preferred: An iterable of preferred types in case
                                   multiple types match.
        :return: The resulting DataType or None if no match was found
        :raises FastrTypeError: if the options argument is of the wrong type

        The function will first create a list of all candidate DataTypes.
        Subsequently, it will check for each candidate if the value would
        valid. If there are multiple matches, the config value for preferred
        types is consulted to break the ties. If non of the DataTypes are in
        the preferred types list, a somewhat random DataType will be picked
        as the most optimal result.
        """
        extra_preferred = None
        log.debug('Guesstype value: {}, options: {}, preferred: {}'.format(value, options, preferred))
        if options is None:
            options = {x for x in self.values() if issubclass(x, DataType)}
        elif issubclass(options, TypeGroup):
            extra_preferred = options.preference
            options = set(options.members)
        elif issubclass(options, DataType):
            options = {options}
        elif isinstance(options, tuple):
            options = set(options)
        else:
            raise exceptions.FastrTypeError('Invalid type for options ({})'.format(options))

        log.debug('Guesstype options: {}'.format(options))
        candidates = set()

        scheme = None
        if url.isurl(value):
            scheme = url.get_url_scheme(value)
            if scheme not in ['vfs', 'val']:
                log.warning('Cannot determine DataType based on URL with scheme {}'.format(url.get_url_scheme(value)))
                return None

        for option in options:
            if not issubclass(option, DataType):
                continue
            elif issubclass(option, Deferred):
                continue  # We never want to find a deferred (it is not a valid type for an instance)
            elif option.dot_extension is None or (isinstance(value, str)
                                                  and value.endswith(option.dot_extension)):
                candidates.add(option)

        log.debug('Guesstype candidates: {}'.format(candidates))
        if len(candidates) == 0:
            log.debug('No valid combinations of options and candidates!')
            return None

        if len(candidates) != 1 and exists:
            # Test validity of value for each DataType
            final_candidates = []
            for candidate in candidates:
                temp = candidate(value)
                if temp.valid:
                    final_candidates.append(candidate)
        else:
            final_candidates = list(candidates)

        # Remove types in these order in case of mutliple matches, that means
        # the Int has precidence over the Boolean, Float and String in case
        # of a tie
        types_to_remove = ['String', 'Float', 'Boolean', 'Int', 'UnsignedInt']

        log.debug('Final candidates: {}'.format(final_candidates))

        if len(final_candidates) > 1:
            for type_to_remove in types_to_remove:
                if types[type_to_remove] in final_candidates:
                    final_candidates.remove(types[type_to_remove])
                if len(final_candidates) == 1:
                    break

        if len(final_candidates) == 0:
            return None
        elif len(final_candidates) == 1:
            log.debug('Matched a single type: {}'.format(final_candidates[0]))
            return final_candidates[0]
        else:
            log.debug('Multiple DataTypes match, trying to find a preferred match! Remaining candidates: {}'.format(final_candidates))

            # Get preferred types from argument list
            if preferred is not None:
                for type_ in preferred:
                    if type_ in final_candidates:
                        log.info('Found preferred match (from keyword): {}'.format(type_))
                        return type_

            # Get preferred types from argument list
            if extra_preferred is not None:
                for type_ in extra_preferred:
                    if type_ in final_candidates:
                        log.info('Found preferred match (from keyword): {}'.format(type_))
                        return type_

            # Get preferred information from the config
            for type_ in self.preferred_types:
                if type_ in final_candidates:
                    log.debug('Found preferred match: {}'.format(type_))
                    return type_

            # Fall back to possible typegroup preferred types

        log.debug('Mutliple matches, removing matches without extension')
        backup_candidate = final_candidates[0]
        final_candidates = [x for x in final_candidates if x.extension is not None]

        if len(final_candidates) == 1:
            return final_candidates[0]
        else:
            if len(final_candidates) > 1:
                log.error('Multiple DataTypes match, but no preferred match, picking one at random! Remaining candidates: {}'.format(final_candidates))
                return final_candidates[0]
            else:
                log.error('No final DataTypes match value "{}", using the backup type: {}'.format(value,
                                                                                                  backup_candidate))
                return backup_candidate

    def match_types(self, *args, **kwargs):
        """
        Find the match between a list of DataTypes/TypeGroups, see :ref:`resolve-datatype` for details

        :param args: A list of DataType/TypeGroup objects to match
        :param kwargs: A 'preferred' keyword argument can be used to indicate a list of DataTypes to prefer in case of ties (first has precedence over later in list)
        :return: The best DataType match, or None if no match is possible.
        :raises FastrTypeError: if not all args are subclasses of BaseDataType
        """
        options = self.match_types_any(*args)

        # Check if it is a preferred type
        if 'preferred' in kwargs and kwargs['preferred'] is not None:
            if not all([self.isdatatype(item) for item in kwargs['preferred']]):
                message = 'All preferred types must be DataTypes!'
                log.warning(message)

            preferred = kwargs['preferred']
        else:
            preferred = self.preferred_types

        if len(options) == 0:
            log.warning("No matching DataType available (args {})".format(args))
            return None
        elif len(options) == 1:
            # This is a perfect match, no preferences needed
            result = options.pop()

            return result
        else:
            # Get preferred information from the config
            for type_ in preferred:
                if type_ in options:
                    return type_

            # Find a single argument that is a list
            if len(args) == 1 and isinstance(args[0], (list, tuple)):
                args = args[0]

            # Check all typegroups in args for preferred types and use those if possible
            for option in args:
                if not issubclass(option, TypeGroup):
                    continue

                for type_ in option.preference:
                    if type_ in options:
                        return type_

            log.debug("No preferred DataType matches, (options {}, preferred {})".format(options, preferred))
            return None

    def match_types_any(self, *args):
        """
        Find the match between a list of DataTypes/TypeGroups, see :ref:`resolve-datatype` for details

        :param args: A list of DataType/TypeGroup objects to match
        :return: A set with all DataTypes that match.
        :rtype: set
        :raises FastrTypeError: if not all args are subclasses of BaseDataType
        """
        # Find a single argument that is a list
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]

        # Remove typeless str
        args = tuple(arg for arg in args if arg != str)

        if not all([self.isdatatype(item) for item in args]):
            message = 'All arguments must be DataTypes! (Found {})'.format(args)
            log.error(message)
            raise exceptions.FastrTypeError(message)

        # In case there are no args
        if len(args) == 0:
            log.debug("No DataTypes given to match")
            return None

        # Create an initial options set (make sure to copy and not reference the set!)
        if issubclass(args[0], TypeGroup):
            options = set(args[0].members)
        else:
            options = {args[0]}

        # Find intersection of all arguments
        for datatype in args[1:]:
            if isinstance(datatype, DataType):
                datatype = type(datatype)

            if issubclass(datatype, TypeGroup):
                options &= datatype.members
            else:
                options &= {datatype}

        return options

    @staticmethod
    def isdatatype(item):
        """
        Check if item is a valid datatype for the fastr system.

        :param item: item to check
        :return: flag indicating if the item is a fastr datatype
        :rtype: bool
        """
        return isinstance(item, type) and issubclass(item, BaseDataType)


types = DataTypeManager()
